# This file is part of pymadcad,  distributed under license LGPL v3
'''	
	This file is to render floating texts and mesh texts
	
	The current implementation of floating texts uses a bitmap font texture baked at program start.
	It's following the first technique described in this tutorial: https://learnopengl.com/In-Practice/Text-Rendering
		
	For mesh texts, freetype (https://freetype.org/) is used to read the font files and extract curve primitives from it. 
	The primitives are then discretized and triangulated using the madcad functions.
	
	When specified by their name, fonts are searched in the system font directories plus the directories in module variable `fontpath`
'''

import os
import sys
from dataclasses import dataclass

from .common import resourcedir
from .mathutils import (Box, vec3, fvec3, fvec4, vec2, fvec2, fmat4, 
						ceil, sqrt, distance, anglebt, linrange,
						)
from .triangulation import triangulation
from .primitives import Segment
from .mesh import Wire, Web, Mesh, web
from . import settings, rendering

from PIL import Image, ImageFont, ImageDraw
import numpy.core as np
import moderngl as mgl



# TODO: utiliser les methodes et attributs ImgeFont.size, .getmetrics(), etc pour avoir les hauteur et largeur de police

fontpath = [resourcedir]

def font_locations():
	''' yield system and user font directories 
		it looks at module variable `fontpath` and then at system font paths
		
		The function is a generator so if iteration is stopped in its beginning, no more than necessary system call is performed
	'''
	yield from fontpath
	
	home = os.getenv('HOME')
	
	if sys.platform == 'linux':
		yield from exploredir(os.path.join(home, '.fonts'))
		yield from exploredir('/usr/local/share/fonts')
		yield from exploredir('/usr/share/fonts')
		yield '/usr/share/fonts'
	
	elif sys.platform == 'win32':
		yield os.path.join(home, r'AppData\Local\Microsoft\Windows\Fonts')
		yield os.path.join(os.environ['WINDIR'],'fonts')
	
	elif sys.platform == 'darwin':
		yield os.path.join(home, 'Library/Fonts')
		yield '/Library/Fonts'
		yield '/System/Library/Fonts'

def exploredir(dir):
	''' list directory elements as string file paths '''
	if os.path.exists(dir):
		if os.path.isdir(dir):
			yield dir
			for sub in os.listdir(dir):
				yield from exploredir(os.path.join(dir, sub))
	
def font_list():
	''' yield all font files found '''
	for location in font_locations():
		try:
			for name in os.listdir(location):
				if name.endswith('.ttf'):
					yield name
		except (NotADirectoryError, FileNotFoundError):
			pass

def font_path(name):
	''' find the font file path for the given font name 
		if the given name is already a path to an existing font file, return it immediately
	'''
	if os.path.exists(name):
		return name
	
	filename = name+'.ttf'
	for location in font_locations():
		path = os.path.join(location, filename)
		if os.path.exists(path):	return path
	raise FileNotFoundError('unable to find font {}'.format(repr(name)))

def create_font_texture(font: ImageFont, maxchar=1100) -> '(Image, (c,l))':
	''' create a font texture by rasterizing a given font '''
	# determine the size of the needed texture
	fontsize = font.size + 4
	width = 64
	while 2*(width//fontsize)**2 < maxchar:
		width *= 2
	c = width//(fontsize//2)
	l = maxchar//c +1
	width = c * (fontsize//2)
	height = l * fontsize
	# create texture
	tex = Image.new('L', (width,height), 0)
	draw = ImageDraw.Draw(tex)
	# draw the font onto
	for i in range(maxchar):
		draw.text(char_placement(font.size, c, l, i), chr(i), fill=255, font=font)
	#tex.show()
	
	return tex, (c, l)

def char_placement(fontsize, c, l, n):
	fontsize += 4
	return ((fontsize//2) * (n%c), fontsize * (n//c))


class TextDisplay(rendering.Display):
	''' halo display of a monospaced text 
	
		This class is usually used through `scheme.note_floating()`
	'''
	pointsdef = [
		[0, 0],
		[0,-1],
		[1,-1],
		[0, 0],
		[1,-1],
		[1, 0],
		]
	
	def __init__(self, scene, position, text, size=None, color=None, font=None, align=(0,0), layer=0):
		if not color:	color = settings.display['annotation_color']
		if not size:	size = settings.display['view_font_size']
		self.position = fvec3(position)
		self.color = fvec3(color)
		self.size = size
		self.layer = layer
		
		# load font
		fontfile = font_path(font or 'NotoMono-Regular')
		def load(scene):
			img, align = create_font_texture(ImageFont.truetype(fontfile, 2*size))
			return scene.ctx.texture(img.size, 1, img.tobytes()), align
		self.fonttex, self.fontalign = scene.resource(('fontfile', size), load)
		
		# load shader
		def load(scene):
			shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/font.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/font.frag').read(),
						)
			shader['fonttex'].value = 0
			return shader
		self.shader = scene.resource('shader_font', load)
		
		# place triangles
		points = np.zeros((len(self.pointsdef)*len(text), 4), 'f4')
		l = 0
		c = 0
		mc = 0
		for i,char in enumerate(text):
			if char == '\n':
				c = 0
				l += 1
			elif char == '\t':
				c += 4 - c%4	# TODO: use a tab size in settings
			elif char == ' ':
				c += 1
			else:
				n = ord(char)
				#placement = char_placement(2*size, *self.fontalign, n)
				for j,add in enumerate(self.pointsdef):
					points[len(self.pointsdef)*i + j] = [
						add[0]+c, 
						add[1]-l, 
						(n % self.fontalign[0] +add[0]) / self.fontalign[0], 
						(n //self.fontalign[0] -add[1]) / self.fontalign[1],
						]
				c += 1
			if c > mc:
				mc = c
		l += 1		
		align = (processalign(align[0], mc), -processalign(align[1], l))
		points -= [*align, 0, 0]
		self.textsize = (mc,l)
		self.visualsize = (-align[0], -align[1], mc//2-align[0], l-align[1])
			
		
		# create the buffers on the GPU
		self.vb_points = scene.ctx.buffer(points)
		self.vb_faces = scene.ctx.buffer(np.arange(points.shape[0], dtype='u4'))
		self.va = scene.ctx.vertex_array(
			self.shader,
			[(self.vb_points, '2f 2f', 'v_position', 'v_uv')],
			)
	
	def render(self, view):		
		self.shader['layer'] = self.layer
		self.shader['color'].write(fvec3(settings.display['select_color_line']) if self.selected else self.color)
		self.shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.shader['view'].write(view.uniforms['view'])
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['ratio'].value = (
				(self.size-2.5) / view.width()*2,
				(self.size-2.5) / view.height()*4,
				)
		self.fonttex.use(0)
		self.va.render(mgl.TRIANGLES)

	def stack(self, view):
		return ((), 'screen', 2, self.render),

def processalign(align, size):
	''' return the concrete alignment, expanding alignments such as 'left', or 'center' to numeric values '''
	if isinstance(align, str):
		if align in ('left', 'top'):		return 0
		elif align in ('right', 'bottom'):	return size
		elif align == 'center':	return size/2
		else:
			raise ValueError("align must be int, float or any of 'left', 'right', 'center'")
	else:
		return align

def textsize(text, tab=4):
	''' visual size of a text as vec2(column,line) '''
	l = 0
	c = 0
	mc = 0
	for i,char in enumerate(text):
		if char == '\n':
			c = 0
			l += 1
		elif char == '\t':
			c += tab - c%tab
		else:
			c += 1
		if c > mc:
			mc = c
	l += 1
	return fvec2(mc, l)




BezierLinear = Segment

@dataclass
class BezierQuadratic:
	''' primitive representing a quadratic Bezier spline '''
	a: vec3
	b: vec3
	c: vec3
	
	def __call__(self, t):
		u = 1-t
		return u**2*self.a + 2*u*t*self.b + t**2*self.c
		
	def mesh(self, resolution=None):
		div = settings.curve_resolution(
			distance(self.a, self.b) + distance(self.b, self.c),
			anglebt(self.b-self.a, self.c-self.b),
			resolution)
		return Wire(self(t)  for t in linrange(0, 1, div=div))
		
	def display(self, scene):
		return self.mesh().display(scene)
	
@dataclass
class BezierCubic:
	''' primitive representing a cubic Bezier spline '''
	a: vec3
	b: vec3
	c: vec3
	d: vec3
	
	def __call__(self, t):
		u = 1-t
		return u**3*self.a + 3*u**2*t*self.b + 3*u*t**2*self.c + t**3*self.d
		
	def mesh(self, resolution=None):
		div = settings.curve_resolution(
			distance(self.a, self.b) + distance(self.b, self.c) + distance(self.c, self.d),
			anglebt(self.b-self.a, self.c-self.b) + angletbt(self.c-self.b, self.d-self.c),
			resolution)
		return Wire(self(t)  for t in linrange(0, 1, div=div))
		
	def display(self, scene):
		return self.mesh().display(scene)

''' font caching dictionnary 
	for each font in the cache, an item `'file.ttf': dict` is created
	each key in this dictionnary is a string character, giving a dictionnary of already computed stuff about the matchng glyph in the font file
	
	This is an example of its content:
	
		cache_fonts = {
			'monospace.ttf': {
				'cbox': Box(...),
				'web': Web(...),
				'mesh': Mesh(...),
				}
			}
'''
cache_fonts = {}

try:
	import freetype
except ImportError:
	pass
else:
	def character_primitives(face: freetype.Face) -> list:
		''' return a list of primitives describing the outline of the character currently loaded in the given face '''
		primitives = []
		last = [None]

		def move_to(a, ctx):
			last[0] = ft2vec(a)
		def line_to(b, ctx):
			primitives.append(BezierLinear(last[0], ft2vec(b)))
			last[0] = ft2vec(b)
		def conic_to(b, c, ctx):
			primitives.append(BezierQuadratic(last[0], ft2vec(b), ft2vec(c)))
			last[0] = ft2vec(c)
		def cubic_to(b, c, d, ctx):
			primitives.append(BezierCubic(last[0], ft2vec(b), ft2vec(c), ft2vec(d)))
			last[0] = ft2vec(d)
			# in fact there will be no cubic_to because there is no such things in freetype fonts, but just in case for freetype-py
		
		face.glyph.outline.decompose(move_to=move_to, line_to=line_to, conic_to=conic_to, cubic_to=cubic_to)
		return primitives

	def ft2vec(ft):   return vec3(ft.x, ft.y, 0)

	def character_cbox(face: freetype.Face) -> Box:
		''' return the cbox of the current glyph loaded in the face '''
		box = face.glyph.outline.get_cbox()
		return Box(
				min=vec3(box.xMin, box.yMin, 0), 
				max=vec3(box.xMax, box.yMax, 0))
		
	def character_outline(char: str, font: str, resolution=None) -> dict:
		''' return a discretized character outline, cached '''
		try:	
			return cache_fonts[font][char]
		except (KeyError, AttributeError):
			if font not in cache_fonts:			cache_fonts[font] = {}
			if char not in cache_fonts[font]:	cache_fonts[font][char] = {}
			cache = cache_fonts[font][char]
			
			if isinstance(font, str):
				face = freetype.Face(font_path(font))
			else:
				face = font
			scale = 1024
			face.set_char_size(scale)
			face.load_char(char)
			cache['fixed'] = face.is_fixed_width
			cache['cbox'] = character_cbox(face) .transform(1/scale)
			cache['web'] = web(character_primitives(face), resolution=resolution) .transform(1/scale) .mergegroups()
			return cache

	def character_surface(char, font, resolution=None) -> dict:
		''' return a triangulated character, cached '''
		try:	
			return cache_fonts[font][char]
		except (KeyError, AttributeError):
			cache = character_outline(char, font, resolution)
			cache['mesh'] = triangulation(cache['web'])
			return cache

	def text(text: str, font:str=None, size:float=1, spacing=vec2(0.05, 0.2), fill=True, align=(0,0), resolution=None):
		''' return a Mesh/Web containing the given text written using the given font
		
			The meshed font is cached so long texts are still fast to mesh
			
			Parameters:
				text:   a multiline string to represent
				font:   the string name of a font such as `'NotoMono-Bold'` or the path to a `.ttf` font file
				size:   the character size (metric unit)
				spacing:  spacing ratio between characters `vec2(horizontal, vertical)`
				fill:   if True, the characters are triangulated and the function returns a `Mesh`, else it returns its outline as a `Web`
				
				align:  
				
					text alignment, the tuple items can be:
					
					- 'left' or 'top'
					- 'center'
					- 'right' or 'bottom'
					- any float, as an offset on the position
				
				resolution:  discretisation setting for the character primitives
				
			Example:
				
				>>> part = text('Hello everyone.\\nthis is a great font !!', 
				...             font='NotoSans-Regular', 
				...             align=('left', 0),
				...             fill=True)
			
		'''
		if fill:	
			pool = Mesh()
			character = character_surface
		else:
			pool = Web()
			character = character_outline
		
		face = freetype.Face(font_path(font or 'NotoMono-Regular'))
		position = vec3(0)
		for char in text:
			if char == ' ':
				position.x += 0.3
			elif char == '\t':
				width = 0.3
				position.x += int(position.x/width) * width
			elif char == '\n':
				position.y -= (1+spacing.y)
				position.x = 0
			else:
				cache = character(char, face, resolution)
				if fill:	part = cache['mesh']
				else:		part = cache['web']
				pool += part.transform(position-vec3(cache['cbox'].min.x,0,0))
				if cache['fixed']:
					position.x += 0.5 + spacing.x
				else:
					position.x += cache['cbox'].width.x + spacing.x
		
		if size != 1:	
			pool = pool.transform(size)
		if align != (0,0):	
			width = pool.box().width
			pool = pool.transform(vec3(
					processalign(align[0], -width[0]),
					processalign(align[1], -width[1]),
					0))
		return pool




def test_character_primitives():
	from .rendering import show
	
	face = freetype.Face('cad/pymadcad/madcad/NotoMono-Regular.ttf')
	face.set_char_size(1024)
	face.load_char('&')
	show([ triangulation(character_primitives(face)) ])

def test_character_cached():
	from .rendering import show
	show([ 
		character_outline('&', 'NotoMono-Regular').web,
		character_surface('g', 'NotoMono-Regular').mesh.transform(vec3(1,0,0)),
		])

def test_text():
	from .rendering import show
	from .generation import extrusion
	from .scheme import note_distance
	settings.display['view_font_size'] = 10
	part = text('Hello everyone.\nthis is a great font !!', font='NotoSans-Regular', align=('center', 0))
	part = extrusion(part.flip(), vec3(0,0,-1))
	show([
			vec3(0),
			note_distance(vec3(0), vec3(0,1,0), offset=vec3(-0.5,0,0), text='size'),
			part,
			], 
		display_wire=True,
		)



from .mathutils import *
from .mesh import Mesh
from .hashing import connef, edgekey
from .constraints import SolveError
from .nprint import nprint

import numpy as np
from scipy.optimize import minimize, least_squares

__all__ = [	'segmentation',
			'guessjoint', 'guesssurface',
			]


def segmentation(mesh, tolerance=5, sharp=0.2) -> Mesh:
	''' Surface segmentation based on curvature.
		This functions splits faces into groups based on curvature proximity.
		Ideally each resulting group is a set of faces with the same curvature. In practice such is not possible due to the mesh resolution introducing bias in curvature estimation. Thus a `tolerance` is used to decide what is a sufficiently close curvature to belong to the same group.
		In order to avoid too sharp edges to be considered a very low resolution but smooth surface, `sharp` is the value of the limit edge angle that can be considered in smooth surface.
		
		This function is particularly usefull when importing geometries from a format that doesn't manage groups (such as STL).
		
		Parameters:
		
			tolerance (float):	maximum difference factor between curvatures of a same group  (1 means +100% curvature is allowed)
			sharp (float):	angle above which an angle is always sharp (radians)
			
		NOTE:
			
			This functions is not magic and despite the default parameters are usually fine for mechanical parts, you may get badly grouped faces in some cases, hence you will need to tune the parameters. 
			
			The success of this functions is highly dependent of the quality of the input mesh triangulation.
			
		Example:
		
			>>> part = read('myfile.stl')
			>>> part.mergeclose()
			>>> part = segmentation(part)
	'''
	# pick informations from the mesh
	pts = mesh.points
	conn = connef(mesh.faces)
	facenormals = mesh.facenormals()
	prec = 1 / mesh.maxnum()  # curvature precision, curvature is rad/m, so 1rad/max dist seems to be a fairly low curvature
	
	# this is what this functions is working to create: new groups and tracks to assign it to faces
	tracks = [-1] * len(mesh.faces)
	groups = []
	
	# evaluate curvature around every edge
	edgecurve = {}
	facecurve = [0.] * len(mesh.faces)
	weight = [0.] * len(mesh.faces)
	for e, fa in conn.items():
		if edgekey(*e) in edgecurve:	continue
		fb = conn.get((e[1],e[0]))
		if not fb:	continue
		
		curve = dot(pts[e[1]] - pts[e[0]], cross(facenormals[fa], facenormals[fb]))
		edgecurve[edgekey(*e)] = angle = anglebt(facenormals[fa], facenormals[fb])
		
		# report average curvature to neighboring faces
		if sharp > angle:
			w = max(0, 0.5-angle)  # empirical weighting
			facecurve[fa] += curve * w  # multiply angle at edge by   (triangle height) ~= surface/(edge length)
			facecurve[fb] += curve * w
			weight[fa] += w
			weight[fb] += w
	
	# divide contributions to face curvatures
	for i,f in enumerate(mesh.faces):
		area = 0.5 * length(cross(pts[f[1]]-pts[f[0]], pts[f[2]]-pts[f[0]])) * weight[i]
		if area:	facecurve[i] /= area
		else:		facecurve[i] = 0
		#debug.append(text.Text(
							#(pts[f[0]] + pts[f[1]] + pts[f[2]]) /3,
							#'{:.2g}'.format(facecurve[i]),
							#size=8,
							#align=('left', 'center'),
							#))
		
	# groupping gives priority to the less curved faces, so they can extend to the ambiguous limits if there is
	sortedfaces = sorted(range(len(mesh.faces)), key=lambda i:  abs(facecurve[i]))
	index = 0
	
	while True:
		# pick a non assigned face
		for index in range(index, len(sortedfaces)):
			if tracks[sortedfaces[index]] == -1:	break
		if index >= len(sortedfaces):	break
		i = sortedfaces[index]
		f = mesh.faces[i]
		
		# start propagation for this face
		tracks[i] = len(groups)
		estimate = facecurve[i]   # average curvature of the group
		sample = 1   # samples used in the average curvature of the group
		front = [(f[k-1], f[k-2])  for k in range(3)]   # propagation frontline
		reached = {i}  # faces reached by propagation
		
		# for triangles on the frontier of a curved surface, the curvature may appear to be null (or close), in this case look for a paired triangle
		if abs(facecurve[i]) < prec:
			best = None
			fence = inf
			for i,e in enumerate(front):
				j = conn.get(e)
				if j is None or tracks[j] != -1 or j in reached:	continue
				if edgecurve[edgekey(*e)] >= sharp:	continue
				score = abs(facecurve[j]-facecurve[i])
				if score < fence:
					best, fence = j, score
			if best is not None:
				estimate = facecurve[best]
		
		# propagation
		while front:
			e = front.pop()
			j = conn.get(e)
			# check if already processed
			if j is None or tracks[j] != -1 or j in reached:	continue
			reached.add(j)
			# check criterions:
			# split groups at maximum curvature
			if edgecurve[edgekey(*e)] >= sharp:	continue
			# split groups when curvature is too far from average curvature
			if abs(facecurve[j]-estimate) / (abs(estimate)+prec) > tolerance:	continue
			
			# improve estimation of average curvature
			tracks[j] = len(groups)
			estimate = (estimate*sample + facecurve[j]) / (sample+1)
			sample += 1
			
			# propagate
			f = mesh.faces[j]
			front.extend((f[k-1], f[k-2])  for k in range(3))
		
		groups.append(estimate)
		index += 1
			
	return Mesh(mesh.points, mesh.faces, tracks, groups)
	
	
	
# ---------   guess-stuff ----------

def guessjoint(solid1, solid2, surface1, surface2, guess=quat(), precision=1e-5) -> 'joint':
	''' Create a kinematic joint based on the surfaces given. The function will use `guesssurface` to find the surface kind and then deduce the appropriate joint to make the surfaces coincide.
	
		Parameters:
		
			solid1, solid2:  	 solids the joint applies on
			surface1, surface2:  the surface to retreive the joint definition from
			guess(quat, mat3):   orientation tip to orient properly the joint axis, if there is
			precision(float):    precision for `guesssurface`
	'''
	from .joints import Planar, Gliding, Punctiform, Ball
	
	joints = {
		('plane', 'plane'): Planar,
		#('cylinder', 'plane'): Rolling,
		('cylinder', 'cylinder'): Gliding,
		#('cylinder', 'sphere'): Anular,
		('plane', 'sphere'): Punctiform,
		('sphere', 'sphere'): Ball,
		}
	t1, p1 = guesssurface(surface1, precision)
	t2, p2 = guesssurface(surface2, precision)
	joint_type = joints.get(tuple(sorted([t1, t2])))
	if not joint_type:
		raise ValueError('not junction registered for {}-{}'.format(t1, t2))
		
	if joint_type in (Gliding, Planar):
		if dot(guess*p1[1], p2[1]) < 0:
			p1 = p1[0], -p1[1]
	return joint_type(solid1, solid2, p1, p2)

from operator import itemgetter
	
def guesssurface(surface, precision=1e-5, attempt=False):	
	''' Guess the surface kind, returns its parameters.
	
		Return a tuple `('surface_type', parameters)`
		
		Parameters:
		
			surface(Mesh):      mesh from which to find the surface kind
			precision(float):	typical distance error from mesh to ideal surface in the surface regression
			attempt(bool):      if True, will not raise an error if the acheived error is too big
	'''
	if not isinstance(surface, Mesh):
		return surface
	
	center = surface.barycenter()
	scores = []
	
	solverargs = dict(
				ftol = precision, 
				max_nfev = 300, 
				method = 'lm')
	
	# plane regression
	def evaluate(x):
		# retreive surface parameters
		origin, normal = vec3(x[:3]), normalize(vec3(x[3:]))
		# allocate residuals
		residual = np.zeros((len(surface.faces)+2, 4), dtype='f8')
		# compute residuals
		residual[-1,0] = (length(vec3(x[3:])) - 1)**2
		residual[-2,0] = length2(noproject(center-origin, normal))
		for i,f in enumerate(surface.faces):
			fp = surface.facepoints(f)
			p = (fp[0]+fp[1]+fp[2]) / 3
			residual[i,0] = dot(p-origin, normal)**2  # distance to the face center
			for j,p in enumerate(fp):
				residual[i,j+1] = dot(p-origin, normal)**2  # distance to each point
		return residual.ravel()
	
	res = least_squares(evaluate, [*center, 1,1,1], **solverargs)
	scores.append((
			np.mean(res.fun),
			'plane',
			(vec3(res.x[:3]), vec3(res.x[3:])),
			))
	#print('plane', res.nfev, np.mean(res.fun), '\t', list(res.x))
	
	# sphere regression
	def evaluate(x):
		# retreive surface parameters
		origin, radius = vec3(x[:3]), float(x[3])
		# allocate residuals
		residual = np.zeros((len(surface.faces), 4), dtype='f8')
		# compute residuals
		for i,f in enumerate(surface.faces):
			fp = surface.facepoints(f)
			p = (fp[0]+fp[1]+fp[2]) / 3
			residual[i,0] = (distance(p, origin) - radius) **2
			for j,p in enumerate(fp):
				residual[i,j+1] = (distance(p, origin) - radius) **2
		return residual.ravel()
	
	res = least_squares(evaluate, [*center, 1], **solverargs)
	scores.append((
			np.mean(res.fun),
			'sphere',
			vec3(res.x[:3]),
			))
	#print('sphere', res.nfev, np.mean(res.fun), '\t', list(res.x))
		
	# cylinder regression
	def evaluate(x):
		# retreive surface parameters
		origin, direction, radius = vec3(x[:3]), vec3(x[3:6]), float(x[6])
		# allocate residuals
		residual = np.zeros((len(surface.faces)+2, 4), dtype='f8')
		# compute residuals
		residual[-1,0] = (length(vec3(x[3:6])) - 1) **2
		residual[-2,0] = length2(project(center-origin, direction))
		for i,f in enumerate(surface.faces):
			fp = surface.facepoints(f)
			p = (fp[0]+fp[1]+fp[2]) / 3
			residual[i,0] = (length(noproject(p-origin, direction)) - radius) **2
			for j,p in enumerate(fp):
				residual[i,j+1] = (length(noproject(p-origin, direction)) - radius) **2
		return residual.ravel()
	# estimate a first axis to avoid the solver to fall into a local extremum
	# sum contrbutions of curvature at edges
	direction = vec3(0)
	conn = connef(surface.faces)
	for e in conn:
		if (e[1], e[0]) in conn and e[0] < e[1]:
			edge = cross(surface.facenormal(conn[e]), surface.facenormal(conn[(e[1],e[0])]))
			direction += edge * (sign(dot(edge, direction)) or 1)
	# if the initial direction is too weak, then we are sure it's not a cylinder
	if length2(direction):
		if -min(direction) > max(direction):	direction = -direction
		# solve
		res = least_squares(evaluate, [*center, *normalize(direction), 1], **solverargs)
		scores.append((
				np.mean(res.fun),
				'cylinder',
				(vec3(res.x[:3]), vec3(res.x[3:])),
				))
		#print('cylinder', res.nfev, np.mean(res.fun), '\t', list(res.x))
		
	# pick best regression of all
	best = min(scores, key=itemgetter(0))
	if attempt or best[0] > precision:
		raise SolveError('unable to find a suitable surface type')
	
	#print('---', best[1])
	return best[1], best[2]



# replace et lisse les points des groupes (contours puis surfaces)
def homogenize(mesh):
	pass

# découpe les faces et replace les points résultants pour améliorer la résolution des surfaes courbes (groupe par groupe)
def subdivide(mesh, density):
	pass

# retire les points les moins importants a la forme pour réduire la densité
def simplify(mesh, density):
	pass



from .container import *
from .wire import Wire


class Web(NMesh):
	''' set of bipoint edges, used to represent wires
		this definition is very close to the definition of Mesh, but with edges instead of triangles
		
		Note:
		
			a Wire instance can contain non-connex geometries (like many separated outlines, called islands), or even non-linear meshing (like intersecting curves). In the purpose of part design, many functions may need more regular characteristics, so checking methods exists and it is up to the user to ensure the mesh do provide them when calling the demanding functions
		
		Attributes:
			points:     typedlist of vec3 for points
			edges:      typedlist of couples for edges, the couple is oriented (meanings of this depends on the usage)
			tracks:     typedlist of integers giving the group each line belong to
			groups:     custom information for each group
			options:	custom informations for the entire web
	'''
	__slots__ = 'points', 'edges', 'tracks', 'groups', 'options'

	# BEGIN --- special methods ---
	
	def __init__(self, points=None, edges=None, tracks=None, groups=None, options=None):
		self.points = ensure_typedlist(points, vec3)
		self.edges = ensure_typedlist(edges, uvec2)
		self.tracks = ensure_typedlist(tracks or typedlist.full(0, len(self.edges), 'I'), 'I')
		self.groups = groups if groups is not None else [None] * (max(self.tracks, default=-1)+1)
		self.options = options or {}
		
	def __add__(self, other):
		''' return a new mesh concatenating the faces and points of both meshes '''
		if isinstance(other, Web):
			r = Web(
				self.points if self.points is other.points else self.points[:], 
				self.edges[:], 
				self.tracks[:], 
				self.groups if self.groups is other.groups else self.groups[:],
				)
			r.__iadd__(other)
			return r
		else:
			return NotImplemented
			
	def __iadd__(self, other):
		''' append the faces and points of the other mesh '''
		if isinstance(other, Web):
			if self.points is other.points:
				self.edges.extend(other.edges)
			else:
				lp = len(self.points)
				self.points.extend(other.points)
				self.edges.extend(e+lp  for e in other.edges)
			if self.groups is other.groups:
				self.tracks.extend(other.tracks)
			else:
				lt = len(self.groups)
				self.groups.extend(other.groups)
				self.tracks.extend(t+lt   for t in other.tracks)
			return self
		else:
			return NotImplemented
			
	# END BEGIN ----- data management -----
	
	def strippoints(self) -> list:
		''' remove points that are used by no faces, return the reindex list.
			if used is provided, these points will be removed without usage verification
			
			return a table of the reindex made
		'''
		self.points, self.edges, reindex = striplist(self.points, self.edges)
		return reindex
	
	def mergepoints(self, merges) -> 'self':
		''' merge points with the merge dictionnary {src index: dst index}
			merged points are not removed from the buffer.
		'''
		j = 0
		for i,f in enumerate(self.edges):
			f = uvec2(
				merges.get(f[0], f[0]),
				merges.get(f[1], f[1]),
				)
			if not f[0] == f[1]:
				self.edges[j] = f
				self.tracks[j] = self.tracks[i]
				j += 1
		del self.edges[j:]
		del self.tracks[j:]
		return self
			
	
	# END BEGIN ----- mesh checks -------
	
	def isline(self):
		''' true if each point is used at most 2 times by edges '''
		reached = typedlist.full(0, len(self.points), 'I')
		for line in self.edges:
			for p in line:	reached[p] += 1
		for r in reached:
			if r > 2:	return False
		return True
	
	def isloop(self):
		''' true if the wire form a loop '''
		return len(self.extremities()) == 0
	
	def check(self):
		''' check that the internal data references are good (indices and list lengths) '''
		if not (isinstance(self.points, typedlist) and self.points.dtype == vec3):	raise MeshError("points must be a typedlist(dtype=vec3)")
		if not (isinstance(self.edges, typedlist) and self.edges.dtype == uvec2): 	raise MeshError("edges must be in a typedlist(dtype=uvec2)")
		if not (isinstance(self.tracks, typedlist) and self.tracks.dtype == 'I'): 	raise MeshError("tracks must be in a typedlist(dtype='I')")
		l = len(self.points)
		for line in self.edges:
			for p in line:
				if p >= l:	raise MeshError("some indices are greater than the number of points", line, l)
				if p < 0:	raise MeshError("point indices must be positive", line)
			if line[0] == line[1]:	raise MeshError("some edges use the same point multiple times", line)
		if len(self.edges) != len(self.tracks):	raise MeshError("tracks list doesn't match edge list length")
		if max(self.tracks, default=-1) >= len(self.groups): raise MeshError("some line group indices are greater than the number of groups", max(self.tracks, default=-1), len(self.groups))
	
	
	
	# END BEGIN ----- selection methods -------
	
	def groupnear(self, point: vec3) -> int:
		''' return group id if the edge the closest to the given point '''
		return self.tracks[self.edgenear(point)]
	
	def edgenear(self, point: vec3) -> int:
		''' return the index of the closest edge to the given point '''
		return min( range(len(self.edges)),
					key=lambda i: distance_pe(point, self.edgepoints(i)) )
	
	def group(self, quals) -> 'Self':
		''' extract a part of the mesh corresponding to the designated groups.
			
			Groups can be be given in either the following ways:
				- a set of group indices
					
					This can be useful to combine with other functions. However it can be difficult for a user script to keep track of which index correspond to which created group
					
				- an iterable of group qualifiers
				
					This is the best way to designate groups, and is meant to be used in combination with `self.qual()`.
					This mode selects every group having all the input qualifiers
					
			Example:
			
				>>> # create a mesh with only the given groups
				>>> mesh.group({1, 3, 8, 9})   
				<Mesh ...>
				
				>>> # create a mesh with all the groups having the following qualifiers
				>>> mesh.group(['extrusion', 'arc'])   
				<Mesh ...>
		'''
		edges = typedlist(dtype=uvec2)
		tracks = typedlist(dtype='I')
		for i in self.qualified_indices(quals):
			edges.append(self.edges[i])
			tracks.append(self.tracks[i])
		return Web(self.points, edges, tracks, self.groups)
	
	def replace(self, mesh, groups=None) -> 'self':
		''' replace the given groups by the given mesh.
			If groups is not specified, it will take the matching groups (with same index) in the current mesh
		'''
		if groups:
			groups = set(self.qualified_groups(groups))
		else:
			groups = set(mesh.tracks)
		j = 0
		for i,t in enumerate(self.tracks):
			if t not in groups:
				self.edges[j] = self.edges[i]
				self.tracks[j] = t
				j += 1
		del self.edges[j:]
		del self.tracks[j:]
		self += mesh
		return self
	
	# END BEGIN --- extraction methods ---
	
	def edgepoints(self, e) -> tuple:
		''' tuple of the points for edge `e`
		
			the edge can be given using its index in `self.edges` or using a tuple of point idinces
		'''
		if isinstance(e, Integral):	e = self.edges[e]
		return self.points[e[0]], self.points[e[1]]
	
	def edgedirection(self, e) -> vec3:
		''' direction of edge e 
			
			the edge can be given using its index in `self.edges` or using a tuple of point idinces
		'''
		if isinstance(e, Integral):	e = self.edges[e]
		return normalize(self.points[e[1]] - self.points[e[0]])
	
	def flip(self) -> 'Web':
		''' reverse direction of all edges '''
		return Web(	self.points, 
					typedlist((uvec2(b,a)  for a,b in self.edges), dtype=uvec2), 
					self.tracks, 
					self.groups)
		
	def segmented(self, group=None) -> 'Web':
		''' return a copy of the mesh with a group each edge 
		
			if group is specified, it will be the new definition put in each groups
		'''
		return Web(self.points, self.edges,
					typedlist(range(len(self.edges)), dtype='I'),
					[group]*len(self.edges),
					self.options,
					)
	
	def subdivide(self, div=1) -> 'Self':
		''' Subdivide all edges by the number of cuts '''
		n = div+2
		pts = typedlist(dtype=vec3)
		edges = typedlist(dtype=uvec2)
		tracks = typedlist(dtype='I')
		c = 0
		for e,t in enumerate(self.tracks):
			# place the points
			o,p0 = self.edgepoints(e)
			x = p0-o
			for i in range(n):
				u = i/(n-1)	
				p = o + u*x
				pts.append(p)
			# create the edges
			for i in range(n-1):
				s = c+i
				edges.append(uvec2(s, s+1))
			c += n
			tracks.extend([t] * (len(indices)-len(tracks)))
		
		new = Web(pts, edges, tracks, self.groups)
		new.mergeclose()
		return new
	
	def extremities(self) -> set:
		''' return the points that are used once only (so at wire terminations)
			1D equivalent of Mesh.outlines()
		'''
		extr = set()
		for l in self.edges:
			for p in l:
				if p in extr:	extr.remove(p)
				else:			extr.add(p)
		return extr
	
	def groupextremities(self) -> 'Wire':
		''' return the extremities of each group.
			1D equivalent of Mesh.groupoutlines()
			
			On a frontier between multiple groups, there is as many points as groups, each associated to a group.
		'''
		indices = typedlist(dtype='I')
		tracks = []
		tmp = {}
		# insert points belonging to different groups
		for i,edge in enumerate(self.edges):
			track = self.tracks[i]
			for p in edge:
				if p in tmp:
					if tmp[p] != track:
						indices.append(p)
						tracks.append(track)
					del tmp[p]
				else:
					tmp[p] = track
		indices.extend(tmp.keys())
		tracks.extend(tmp.values())
		return Wire(self.points, indices, tracks, self.groups)
		
	def frontiers(self, *args) -> 'Wire':
		''' return a Wire of points that split the given groups appart.
		
			The arguments are groups indices or lists of group qualifiers (as set in `qualify()`). If there is one only argument it is considered as as list of arguments.
		
			- if no argument is given, then return the frontiers between every groups
			- to include the groups extremities that are on the group border but not at the frontier with an other group, add `None` to the group set
			
			Example:
			
				>>> w = Web([...], [uvec2(0,1), uvec2(1,2)], [0, 1], [...])
				>>> w.frontiers(0,1).indices
				[1]
				
				>>> # equivalent to
				>>> w.frontiers({0,1}).indices
				[1]
				
				>>> w.frontiers(0,None).indices
				[0]
		'''
		if args:
			if len(args) == 1 and hasattr(args[0], '__iter__'):
				args = args[0]
			groups = set()
			for arg in args:
				if arg is None:		groups.add(None)
				else:				groups.update(self.qualified_groups(arg))
		else:
			groups = None
		
		indices = typedlist(dtype='I')
		tracks = typedlist(dtype='I')
		couples = OrderedDict()
		belong = {}
		for i,edge in enumerate(self.edges):
			track = self.tracks[i]
			if groups is None or None in groups or track in groups:	
				for p in edge:
					if p in belong:
						if belong[p] != track and (groups is None or track in groups and belong[p] in groups):
							g = edgekey(belong[p],track)
							indices.append(p)
							tracks.append(couples.setdefault(g, len(couples)))
						del belong[p]
					else:
						belong[p] = track
		if groups and None in groups:
			for p, i in belong.items():
				if i in groups:
					g = (i,None)
					indices.append(p)
					tracks.append(couples.setdefault(g, len(couples)))
		return Wire(self.points, indices, tracks, list(couples))
		
		
	
	def length(self) -> float:
		''' total length of edges '''
		s = 0
		for e in self.edges:
			s += distance(self.points[e[1]], self.points[e[0]])
		return s
	
	def surface(self) -> float:
		''' return the surface enclosed by the web if planar and is composed of loops (else it has no meaning) '''
		o = self.barycenter()
		s = vec3(0)
		for e in self.edges:
			s += cross(self.points[e[1]] - o, self.points[e[0]] - o)
		return length(s)
	
	def barycenter(self) -> vec3:
		''' curve barycenter of the mesh '''
		if not self.edges:	return vec3(0)
		acc = vec3(0)
		tot = 0
		for e in self.edges:
			a,b = self.edgepoints(e)
			weight = distance(a,b)
			tot += weight
			acc += weight*(a+b)
		return acc / (2*tot)
		
	def assignislands(self) -> 'Web':
		conn = connpe(self.edges)
		reached = [False] * len(self.edges)
		stack = []
		start = 0
		group = 0
		while True:
			# search start point
			for start in range(start,len(reached)):
				if not reached[start]:
					stack.append(start)
					break
			# end when everything reached
			if not stack:	break
			# propagate
			island = Web(points=self.points, groups=self.groups)
			while stack:
				i = stack.pop()
				if reached[i]:	continue	# make sure this face has not been stacked twice
				reached[i] = True
				yield i, group
				for p in self.edges[i]:
					stack.extend(n	for n in conn[p] if not reached[n])
			group += 1
	
	def groupislands(self) -> 'Web':
		''' return the same web but with a new group each island '''
		tracks = typedlist.full(0, len(self.edges), dtype='I')
		track = 0
		for edge, track in self.assignislands():
			tracks[edge] = track
		return Web(self.points, self.edges, tracks, [None]*(track+1))
		
	def islands(self) -> '[Web]':
		''' return the unconnected parts of the mesh as several meshes '''
		islands = []
		island = Web(points=self.points, groups=self.groups)
		for edge, group in self.assignislands():
			if group > len(islands):
				islands.append(island)
				island = Web(points=self.points, groups=self.groups)
			else:
				island.edges.append(self.edges[edge])
				island.tracks.append(self.tracks[edge])
		islands.append(island)
		return islands
	
	def arcs(self) -> '[Wire]':
		''' return the contiguous portions of this web '''
		return [	Wire(self.points, typedlist(loop, dtype=uvec2))		
					for loop in suites(self.edges, oriented=False)]
		
	# END BEGIN ----- output methods ------
	
	def __repr__(self):
		return '<Web with {} points at 0x{:x}, {} edges>'.format(len(self.points), id(self.points), len(self.edges))
	
	def __str__(self):
		return 'Web(\n  points={},\n  edges={},\n  tracks={},\n  groups={},\n  options={})'.format(
					reprarray(self.points, 'points'),
					reprarray(self.edges, 'edges'),
					reprarray(self.tracks, 'tracks'),
					reprarray(self.groups, 'groups'),
					repr(self.options))
					
	def display(self, scene):		
		from .. import displays
		
		points = typedlist(dtype=vec3)
		idents = typedlist(dtype='I')
		edges = typedlist(dtype=uvec2)
		frontiers = typedlist(dtype='I')
		def usept(pi, ident, used):
			if used[pi] >= 0:	
				return used[pi]
			else:
				used[pi] = i = len(points)
				points.append(self.points[pi])
				idents.append(ident)
				return i
		
		for group in range(len(self.groups)):
			used = [-1]*len(self.points)
			frontier = set()
			for edge,track in zip(self.edges, self.tracks):
				if track != group:	continue
				edges.append(uvec2(usept(edge[0], track, used), usept(edge[1], track, used)))
				for p in edge:
					if p in frontier:	frontier.remove(p)
					else:				frontier.add(p)
			for p in frontier:
				frontiers.append(used[p])
				
		if not points or not edges:
			return displays.Display()
		
		return displays.WebDisplay(scene,
				typedlist_to_numpy(points, 'f4'), 
				typedlist_to_numpy(edges, 'u4'),
				typedlist_to_numpy(frontiers, 'u4'),
				typedlist_to_numpy(idents, 'u4'),
				color=self.options.get('color'))
				
	# END





# This file is part of pymadcad,  distributed under license LGPL v3

''' This modules defines the constraints definitions and the solver tools

	Constraints can be any object referencing `variables` and implementing the following signature to guide the solver in the resolution:
		
		class SomeConstraint:
			
			# name the attributes referencing the solver variables to change
			slvvars = 'some_primitive', 'some_point'
			# function returning the squared error in the constraint
			
			#  for a coincident constraint for instance it's the squared distance
			#  the error must be a contiguous function of the parameters, and it's squared for numeric stability reasons.
			#  the solver internally use an iterative approach to optimize the sum of all fit functions.
			def fit((self):
				...
'''

from collections import Counter
import numpy as np
from scipy.optimize import least_squares
from .nprint import nprint
from .mathutils import *
from . import primitives
from . import displays, text, settings
from . import scheme
import array


class SolveError(Exception):	pass

class Constraint(object):
	def __init__(self, *args, **kwargs):
		for i,name in enumerate(self.__slots__):
			setattr(self, name, args[i] if i < len(args) else None)
		for name, arg in kwargs.items():
			setattr(self, name, arg)
	#def fitgrad(self):
		#varset = Varset()
		#for name in self.primitives:
			#varset.register(getattr(self, name))
		#return Derived(varset.state(), varset.grad(), varset.vars)

def isconstraint(obj):
	''' Return True if obj match the constraint signature '''
	return hasattr(obj, 'fit') and hasattr(obj, 'slvvars')

	
	


class Tangent(Constraint):
	''' Makes to curves tangent in the given point 
		The point moves as well as curves.
		
		The curves primitives must have a member ``slv_tangent(p)`` returning the tangent vector at the nearest position to `p`
	'''
	__slots__ = 'c1', 'c2', 'p', 'size'
	slvvars = 'c1', 'c2', 'p'
	def fit(self):
		return length2(cross(self.c1.slv_tangent(self.p), self.c2.slv_tangent(self.p))),
	
	#def display(self, scene):
		#return displays.TangentDisplay(scene, (self.p, self.c2.slv_tangent(self.p)), self.size)

class Distance(Constraint):
	''' Makes two points distant of the fixed given distance '''
	__slots__ = 'p1', 'p2', 'd', 'along', 'location'
	slvvars = 'p1', 'p2'
	def fit(self):
		if isinstance(self.p1, vec3) and isinstance(self.p2, vec3):
			if self.along:
				if isinstance(along, vec3):	a = along
				else:						a = along.direction
				return (dot(self.p1-self.p2, a) - d) ** 2,
			else:
				return (distance(self.p1, self.p2) - self.d) **2,
		elif isinstance(self.p1, vec3):
			return (length(noproject(self.p2.origin-self.p1, self.p2.direction)) - self.d) **2,
		elif isinstance(self.p2, vec3):
			return (length(noproject(self.p1.origin-self.p2, self.p1.direction)) - self.d) **2,
		else:
			d1 = self.p1.direction
			d2 = self.p2.direction
			if dot(d1,d2) < 0:	d2 = -d2
			return length2(cross(d1,d2)), (length(noproject(self.p1.origin-self.p2.origin, d1+d2)) - self.d) **2
			
	#def fitgrad(self):
		#return derived.compose(dot, Derived(self.p1), Derived(self.p2))
	def display(self, scene):
		if isinstance(self.p1, vec3) and isinstance(self.p2, vec3):
			return scene.display(scheme.note_distance(
					self.p1, self.p2, 
					project=self.along,
					text='{:.5g}\n{:+.1g}'.format(self.d, sqrt(self.fit()[0])),
					))
		elif isinstance(self.p1, vec3):
			return scene.display(scheme.note_distance(
					self.p1, self.p1 + noproject(self.p2.origin-self.p1, self.p2.direction),
					text='{:.5g}\n{:+.1g}'.format(self.d, sqrt(self.fit()[0])),
					))
		elif isinstance(self.p2, vec3):
			return scene.display(scheme.note_distance(
					self.p2, self.p2 + noproject(self.p1.origin-self.p2, self.p1.direction), 
					text='{:.5g}\n{:+.1g}'.format(self.d, sqrt(self.fit()[0])),
					))
		else:
			p = mix(self.p1.origin, self.p2.origin, 0.5)
			d1 = self.p1.direction
			d2 = self.p2.direction
			if dot(d1,d2) < 0:	d2 = -d2
			d = d1 + d2
			p1 = self.p1.origin
			p2 = self.p2.origin
			return scene.display(scheme.note_distance(
					p + noproject(p1-p, d),
					p + noproject(p2-p, d),
					text='{:.5g}\n{:+.1g}'.format(self.d, sqrt(self.fit()[0])),
					))
			

class Angle(Constraint):
	''' Gets two segments with the given fixed angle between them '''
	__slots__ = 's1', 's2', 'angle'
	slvvars = 's1', 's2'
	def fit(self):
		d1 = self.s1.direction
		d2 = self.s2.direction
		a = atan2(length(cross(d1,d2)), dot(d1,d2))
		return (a - self.angle)**2,
	
	def display(self, scene):
		return scene.display(scheme.note_angle(
					(self.s1.origin, -self.s1.direction),
					(self.s2.origin, -self.s2.direction),
					text='{:.5g}°\n{:+.1g}'.format(degrees(self.angle), degrees(sqrt(self.fit()[0]))),
					))

class Parallel(Constraint):
	''' Strict equivalent of Angle(s1,s2,0) '''
	__slots__ = 's1', 's2'
	slvvars = __slots__
	def fit(self):
		d1 = self.s1.direction
		d2 = self.s2.direction
		return length2(cross(d1,d2)),
		

class Radius(Constraint):
	''' Gets the given Arc with the given fixed radius 
		
		Note: Only ArcCentered are supported yet.
	'''
	__slots__ = 'arc', 'radius', 'location'
	slvvars = 'arc',	
	def fit(self):
		return (self.arc.radius - self.radius) **2,
	
	def display(self, scene):
		r = self.arc.radius
		center, z = self.arc.axis
		x = dirbase(z)[0]
		return scene.display(scheme.note_leading(
					center+x*r, 
					r*x, 
					text='R{:.5g}\n{:+.1g}'.format(r, self.radius - self.arc.radius),
					))

class OnPlane(Constraint):
	''' Puts the given points on the fixed plane given by its normal axis '''
	__slots__ = 'axis', 'pts'
	def slvvars(self):
		return self.pts
	def fit(self):
		s = 0
		for p in self.pts:
			yield dot(p-self.axis[0], self.axis[1]) **2

class PointOn(Constraint):
	''' Puts the given point on the curve.
	
		The curve primitive must have a member  ``slv_nearest(p) -> vec3`` returning the closest point to p on the curve.
	'''
	__slots__ = 'point', 'curve'	
	slvvars = 'point', 'curve'
	def fit(self):
		return distance2(self.curve.slv_nearest(self.point), self.point),

		
		

def solve(constraints, fixed=(), *args, **kwargs):
	''' Short hand to use the class Problem '''
	return Problem(constraints, fixed).solve(*args, **kwargs)

class Problem:
	''' Class to holds data for a problem solving process.
		it is intended to be instantiated for each different probleme solving, an instance is used multiple times only when we want to solve on top of the previous results, using exactly the same probleme definition (constraints and variables)
		
		Therefore the solver protocol is the following:
			- constraints define the problem
			- each constraint refers to variables it applies on
				constraints have the method fit() and a member 'slvvars' that can be  
				
				1. An iterable of names of variable members in the constraint object
				2. A function returning an iterable of the actual variables objects (that therefore must be referenced refs and not primitive types)
			
			- each variable object can redirect to other variable objects if they implements such a member 'slvvars'
			- primitives can also be constraints on their variables, thus they must have a method fit()   (but no member 'primitives' here)
			- primitives can implement the optional solver methods for some constraints, such as 'slv_tangent'
	'''
	def __init__(self, constraints, fixed=()):
		self.constraints = set()
		self.slvvars = {}
		self.dim = 0
		for cst in constraints:
			# cst can contains objects that are not constraints
			if hasattr(cst, 'fit'):
				self.register(cst)
		for prim in fixed:
			self.unregister(prim)
		for v in self.slvvars.values():
			if isinstance(v, tuple):	self.dim += 1
			else:						self.dim += len(v)
	
	def register(self, obj):
		''' Register a constraint or a variable object '''
		if hasattr(obj, 'fit'):
			self.constraints.add(obj.fit)
		# register object's variables
		if hasattr(obj, 'slvvars'):
			if callable(obj.slvvars):
				for var in obj.slvvars():
					if isinstance(var, (float, int)):		raise TypeError("primitive types (float,int) are not allowed when 'slvvars' is a callable")
					self.register(var)
			else:
				for varname in obj.slvvars:
					var = getattr(obj, varname)
					if isinstance(var, (float,int)):
						self.slvvars[(id(obj), varname)] = (obj, varname)
					else:
						self.register(var)
		elif isinstance(obj, tuple):
			for p in obj:
				self.register(p)
		else:
			if id(obj) not in self.slvvars:
				self.slvvars[id(obj)] = obj
	
	def unregister(self, obj):
		''' Unregister all variables from a constraint or a variable object '''
		if hasattr(obj, 'slvvars'):
			if callable(obj.slvvars):
				for var in obj.slvvars():
					self.unregister(var)
			for varname in obj.slvvars:
				var = getattr(obj, varname)
				if isinstance(var, (float,int)):
					del self.slvvars[(id(obj), varname)]
				else:
					self.unregister(var)
		elif isinstance(obj, tuple):
			for p in obj:
				self.unregister(p)
		else:
			if id(obj) in self.slvvars:
				del self.slvvars[id(obj)]
	
	def state(self):
		x = np.empty(self.dim, dtype='f8')
		i = 0
		for v in self.slvvars.values():
			l = len(v)
			x[i:i+l] = v
			i += l
		return x
	
	def place(self, x):
		i = 0
		for v in self.slvvars.values():
			l = len(v)
			for j in range(l):	v[j] = x[i+j]
			i += l
	
	def fit(self):
		residuals = array.array('d')
		for fit in self.constraints:
			residuals.extend(fit())
		return residuals
	
	def evaluate(self, x):
		self.place(x)
		return self.fit()
	
	def solve(self, precision=1e-6, method='trf', maxiter=None, afterset=None):
		#nprint(self.slvvars)
		if afterset:	evaluate = lambda x: afterset(x) or self.evaluate(x)
		else:			evaluate = self.evaluate
		res = least_squares(evaluate, self.state(), 
				xtol=precision, 
				gtol=precision**3,
				ftol=precision**2,
				method=method, 
				max_nfev=maxiter,
				)
		
		self.place(res.x)
		#print(res)
		if res.cost < precision:
			return res
		elif not res.sucess:
			raise SolveError(res.message)
		else:
			raise SolveError('no solution found')




def solve2(constraints, precision=1e-4, afterset=None, fixed=(), maxiter=0):
	params = []
	corrections = []
	corrnorms = []
	
	indices = []
	knownparams = {}
	
	parts = []
	
	mincorr = precision * 1e-2
	
	# get parameters and prepare solver
	i = 0
	for const in constraints:
		c = 0
		for param in const.params():
			k = id(param)
			if k not in fixed:	c += 1
			if k not in knownparams:
				knownparams[k] = i
				params.append(param)
				corrections.append(None)
				corrnorms.append(0)
				i += 1
			indices.append(knownparams[k])
		parts.append(c)
	
	# iterative resolution
	oldcorrs = [None] * len(corrections)
	maxdelta = inf
	it = 0
	while maxdelta > precision:
		maxcorr = 0
		maxdelta = 0
		
		# initialize corrections
		for i in range(len(corrections)):
			corrections[i] = type(params[i])()
			corrnorms[i] = 0.
		
		# compute constraints contributions to corrections
		i = 0
		for const,part in zip(constraints, parts):
			corr = const.corrections()
			for correction in corr:
				contrib = correction / part
				corrections[indices[i]] += contrib
				l = length(contrib)
				if l > corrnorms[indices[i]]:	corrnorms[indices[i]] = l
				i += 1
		
		# apply changes
		for param,correction,corrnorm,oldcorr in zip(params, corrections, corrnorms, oldcorrs):
			if id(param) not in fixed:
				l = length(correction)
				if l > maxcorr:		maxcorr = l
				if corrnorm > maxdelta:		maxdelta = corrnorm
				# filtre a oscillation
				if oldcorr:
					#print('       ', correction*0.5 + oldcorr*0.5)
					param += (correction*0.5 + oldcorr*0.5)
				else:
					param += correction * 0.8
		oldcorrs = corrections[:]
		
		if afterset:	afterset()
		it += 1
		# check that the solver is solving
		if maxiter and it > maxiter:	
			raise SolveError('failed to converge with the allowed iteration count: '+str(maxiter))
		if maxdelta > precision and maxcorr < mincorr:	
			raise SolveError('resolution is blocked (try an other initial state)')

	#print('iterations:', it-1)
	return maxdelta



#!/usr/bin/python3

from setuptools import setup, find_packages, Extension

try:
	from Cython.Build import cythonize
except ImportError:
	cython_modules = [Extension('madcad.core', ['madcad/core.c'])]
else:
	cython_modules = cythonize(['madcad/core.pyx'])	#, annotate=True)

setup(
	# package declaration
	name='pymadcad',
	version='0.18.1',
	python_requires='>=3.8',
	install_requires=[
		'pyglm>=2.5.5',
		'moderngl>=5.6',
		'numpy>=1.1',
		'scipy>=1.3',
		'PyQt5>=5',
		'Pillow>=5.4',
		'pyyaml>=5',
		'arrex>=0.5.1',
		],
	extras_require={
		'PLY': ['plyfile>=0.7'],
		'STL': ['numpy-stl>=2'],
		'OBJ': ['PyWavefront>=1.3'],
		'freetype': ['freetype-py>=2.3'],
		},
	# source declaration
	packages=find_packages(),
	ext_modules=cython_modules,
	package_data={
		'madcad': [
			'shaders/*.frag', 
			'shaders/*.vert', 
			'textures/*.png', 
			'*.py', 
			'*.pyx', 
			'*.c', 
			'*.ttf',
			],
		'': ['COPYING', 'COPYING.LESSER', 'README.md', 'requirements.txt'],
		},
	
	# metadata for pypi
	author='Yves Dejonghe',
	author_email='jimy.byerley@gmail.com',
	description="Simple yet powerful CAD (Computer Aided Design) library, written with Python",
	long_description=open('README.md', encoding='utf-8').read(),
	long_description_content_type='text/markdown',
	license='GNU LGPL v3',
	url='https://github.com/jimy-byerley/pymadcad',
	keywords='CAD 3D parametric mesh kinematic solid solver part design',
	classifiers=[
		'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
		'Topic :: Scientific/Engineering',
		'Development Status :: 3 - Alpha',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'Programming Language :: Python :: 3.11',
		'Programming Language :: Python :: 3.12',
		'Programming Language :: Python :: Implementation :: CPython',
		'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
		'Intended Audience :: Science/Research',
		'Intended Audience :: Manufacturing',
		'Intended Audience :: Education',
		'Topic :: Scientific/Engineering',
		'Topic :: Scientific/Engineering :: Visualization',
		'Environment :: X11 Applications :: Qt',
		],
	)



# This file is part of pymadcad,  distributed under license LGPL v3

import numpy as np
import numpy.lib.recfunctions as rfn
import os, tempfile
from functools import wraps
from hashlib import md5

from .mathutils import vec3, glm, inf, typedlist
from .mesh import Mesh, Wire

class FileFormatError(Exception):	pass


def filetype(name, type=None):
	''' Get the name for the file format, using the given forced type or the name extension '''
	if not type:
		type = name[name.rfind('.')+1:]
	if not type:
		raise FileFormatError('unable to guess the file type')
	return type
	
def read(name: str, type=None, **opts) -> Mesh:
	''' Load a mesh from a file, guessing its file type '''
	type = filetype(name, type)
	reader = globals().get(type+'_read')
	if reader:
		return reader(name, **opts)
	else:
		raise FileFormatError('no read function available for format '+type)

def write(mesh: Mesh, name: str, type=None, **opts):
	''' Write a mesh to a file, guessing its file type '''
	type = filetype(name, type)
	writer = globals().get(type+'_write')
	if writer:
		return writer(mesh, name, **opts)
	else:
		raise FileFormatError('no write function available for format '+type)

cachedir = tempfile.gettempdir() + '/madcad-cache'
caches = {}
def cache(filename: str, create: callable=None, name=None, storage=None, **opts) -> Mesh:
	''' Small cachefile system, it allows to dump objects to files and to get them when needed.
		
		It's particularly useful when working with other processes. The cached files are reloaded only when the cache files are newer than the memory cache data.
		
		If specified, create() is called to provide the data, in case it doesn't exist in memory neither as file.
		If specified, name is the cache name used to index the file it defaults to the `filename`.
		If specified, storage is the dictionary used to storage cache data, defaults to io.caches.
	'''
	if not storage:	storage = caches
	if not name:	name = filename
	
	# create the cache file if it doesn't exist
	if os.path.exists(filename):
		cachedate = storage[name][0] if name in storage else -inf
		filedate = os.path.getmtime(filename)
		if cachedate < filedate:
			storage[name] = (filedate, read(filename, **opts))
	# load reload the file content if it's newer that the data in memory
	else:
		if name in storage:	obj = storage[name][1]
		elif create:		obj = create()
		else:
			raise IOError("the cache file doesn't exist")
		write(obj, filename, **opts)
		storage[name] = (os.path.getmtime(filename), obj)
	return storage[name][1]
	
	

def cachefunc(f):
	''' Decorator to cache a function results.
		
		Use it if you want to cache their result associated with the argument set used 
	'''
	@wraps(f)
	def repl(*args, **kwargs):
		if not os.path.exists(cachedir):
			os.makedirs(cachedir)
		key = '{}/{}{}-{}.pickle'.format(
			cachedir,
			f.__module__ + '.' if f.__module__ else '',
			f.__name__,
			hex(int.from_bytes(
				md5(repr((
					*args, 
					sorted(list(kwargs.items()))
					)).encode())
					.digest(),
				'little')),
			)
		return cache(key, lambda: f(*args, **kwargs))
	return repl

	
'''
	pickle files are the standard python serialized files, they are absolutely not secure ! so do not use it for something else than your own caching.
'''
import pickle

def pickle_read(file, **opts):
	return pickle.load(open(file, 'rb'))
	
def pickle_write(obj, file, **opts):
	return pickle.dump(obj, open(file, 'wb'))

'''
	PLY is loaded using plyfile module 	https://github.com/dranjan/python-plyfile
	using the specifications from 	https://web.archive.org/web/20161221115231/http://www.cs.virginia.edu/~gfx/Courses/2001/Advanced.spring.01/plylib/Ply.txt
		(also locally available in ply-description.txt)
'''
try:
	from plyfile import PlyData, PlyElement
except ImportError:	pass
else:
	from . import triangulation

	def ply_read(file, **opts):
		mesh = Mesh()
		
		data = PlyData.read(file)
		if 'vertex' not in data:	raise FileFormatError('file must have a vertex buffer')
		if 'face' not in data:		raise FileFormatError('file must have a face buffer')
		
		# collect points
		mesh.points = typedlist(data['vertex'].data.astype('f8, f8, f8'), dtype=vec3)
		
		# collect faces
		faces = data['face'].data
		if faces.dtype.names[0] == 'vertex_indices':
			for face in faces['vertex_indices']:
				#print('  ', type(face), face, face.dtype, face.strides)
				if len(face) == 3:	# triangle
					mesh.faces.append(face)
				elif len(face) > 3:	# quad or other extended face
					mesh += triangulation.triangulation_outline(Wire(mesh.points, face))
		else:
			mesh.faces = numpy_to_typedlist(faces.astype('u4'), dtype=uvec3)

		# collect tracks
		if 'group' in faces.dtype.names:
			mesh.tracks = typedlist(faces['group'].astype('u4'), dtype='I')
		else:
			mesh.tracks = typedlist.full(0, len(mesh.faces), 'I')
		
		# create groups  (TODO find a way to get it from the file, PLY doesn't support non-scalar types)
		mesh.groups = [None] * (max(mesh.tracks, default=-1)+1)
		
		return mesh

	def ply_write(mesh, file, **opts):
		vertices = np.asarray(mesh.points).astype(np.dtype([('x', 'f4'), ('y', 'f4'), ('z', 'f4')]))
		faces = np.empty(len(mesh.faces), dtype=[('vertex_indices', 'u4', (3,)), ('group', 'u2')])
		faces['vertex_indices'] = typedlist_to_numpy(mesh.faces, 'u4')
		faces['group'] = typedlist_to_numpy(mesh.tracks, 'u4')
		ev = PlyElement.describe(vertices, 'vertex')
		ef = PlyElement.describe(faces, 'face')
		PlyData([ev,ef], opts.get('text', False)).write(file)


'''
	STL is loaded using numpy-stl module 	https://github.com/WoLpH/numpy-stl
'''
try:	
	import stl
except ImportError:	pass
else:

	from .mathutils import *
	from .mesh import numpy_to_typedlist, typedlist_to_numpy
	def stl_read(file, **opts):
		stlmesh = stl.mesh.Mesh.from_file(file, calculate_normals=False)
		trinum = stlmesh.points.shape[0]
		mesh = Mesh(
			numpy_to_typedlist(stlmesh.points.reshape(trinum*3, 3), vec3), 
			typedlist(uvec3(i, i+1, i+2)  for i in range(0, 3*trinum, 3)),
			)
		mesh.options['name'] = stlmesh.name
		return mesh

	def stl_write(mesh, file, **opts):
		stlmesh = stl.mesh.Mesh(np.zeros(len(mesh.faces), dtype=stl.mesh.Mesh.dtype), name=mesh.options.get('name'))
		stlmesh.vectors[:] = typedlist_to_numpy(mesh.points, 'f4')[typedlist_to_numpy(mesh.faces, 'i4')]
		stlmesh.save(file, **opts)

'''
	OBJ is loaded using the pywavefront module	https://github.com/pywavefront/PyWavefront
	using the specifications from 	https://en.wikipedia.org/wiki/Wavefront_.obj_file
'''
try:
	import pywavefront
except ImportError:	pass
else:
	
	def obj_read(file, **opts):
		scene = pywavefront.Wavefront(file, parse=True, collect_faces=True)
		points = [vec3(v[:3]) for v in scene.vertices]
		faces = []
		for sub in scene.meshes.values():
			faces.extend(( tuple(f[:3]) for f in sub.faces ))
		mesh = Mesh(points, faces)
		if len(scene.meshes) == 1:
			mesh.options['name'] = next(iter(scene.meshes))
		return mesh
	
	# no write function available at this time
	#def obj_write(mesh, file, **opts):

'''
	JSON is loaded using the builtin json module
	always using the official json specifications
	it can store many object types, not only shapes
'''
import json

class JSONEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, (vec2,vec3,vec4,mat2,mat3,mat4,quat)):
			return {'type':type(obj).__name__, 'content':list(obj)}
		elif isinstance(obj, np.ndarray):
			return {'type':'ndarray', 'dtype':obj.dtype, 'content':list(obj)}
		elif isinstance(obj, Mesh):
			return {'type':'Mesh', 'points': [tuple(p) for p in obj.points], 'faces':obj.faces, 'tracks':obj.tracks, 'groups':obj.groups}
		elif isinstance(obj, Web):
			return {'type':'Web', 'points': [tuple(p) for p in obj.points], 'edges':obj.edges, 'tracks':obj.tracks, 'groups':obj.groups}
		else:
			return json.JSONEncoder.default(self, obj)

def jsondecode(obj):
	if 'type' in obj:
		t = obj['type']
		if t in {'vec2','vec3','vec4','mat2','mat3','mat4','quat'}:		
			return vec3(obj['content'])
		elif t == 'ndarray':
			return np.array(obj['content'], dtype=obj['dtype'])
		elif t == 'Mesh':
			return Mesh([vec3(p) for p in obj['points']], [tuple(f) for f in obj['faces']], obj['tracks'], obj['groups'])
		elif t == 'Web':
			return Mesh([vec3(p) for p in obj['points']], [tuple(f) for f in obj['edges']], obj['tracks'], obj['groups'])
		else:
			raise FileFormatError('unable to load json for dumped type {}', t)
	return obj
	
def json_read(file, **opts):
	return json.load(open(file, 'r'), cls=JSONDecoder, **opts)

def json_write(objs, file, **opts):
	return json.dump(open(file, 'w'), object_hook=jsondecode, **opts)



# This file is part of pymadcad,  distributed under license LGPL v3
'''
	The functions here generare meshes from simpler objects (like lower dimension meshes, or simple primitives). You won't find here iterative methods, nor adaptative geometry operations, nor mesh modification operations.
	The functions here may eventually be complex but will always feel predictable and simple from a human perspective.
'''

from .mathutils import *
from .mesh import Mesh, Web, Wire, MeshError, web, wire, mkquad, mktri
from .hashing import edgekey, facekeyo, suites
from . import triangulation
from . import settings
from . import primitives
from .nprint import nprint

import itertools
from copy import copy


__all__ = [
	'extrans', 'extrusion', 'revolution', 'helix', 'screw', 'saddle', 'tube', 
	'repeat', 'repeataround',
	'flatsurface', 'icosurface',
	'square', 'brick', 'parallelogram', 'cylinder', 'cone', 'pyramid', 'icosahedron', 'icosphere', 'uvsphere', 'regon', 
	]

	
# --- extrusion things ---


def extrusion(shape, trans:transform, alignment:float=0) -> Mesh:
	''' Create a surface by extruding the given outline by a transformation
		
		Parameters:
			shape:         a line (Web or Wire) or a surface (Mesh) to extrude
			trans:        any transformation object accepted by `mathutils.transform`
			alignment:  the relative position of the input shape in the resulting mesh
				- `0` means at the beginning
				- `1` means at the end
				
		Example:
		
			>>> extrusion(ArcThrough(+Y, Z, -Y), 2*X)
	'''
	trans = transform(trans)
	neutral = mat4()
	return extrans(shape, [
				neutral + (trans-neutral)*(-alignment),
				neutral + (trans-neutral)*(1-alignment),
				],
				[(0,1,0)])

def revolution(shape, axis=Axis(O,Z), angle:float=2*pi, alignment:float=0, resolution=None) -> Mesh:
	''' Create a revolution surface by extruding the given outline
		`steps` is the number of steps between the start and the end of the extrusion
		
		Parameters:
			shape:    the shape to extrude (Web, Wire, or Mesh), ideally a section
			angle:    angle of rotation between the given profile and the final produced profile
			axis:     the axis to rotate around
			alignment:  the relative position of the input shape in the resulting mesh
				- `0` means at the beginning
				- `1` means at the end
			resolution:   resolution setting for the biggest produced circle, such as for primitives
			
		Example:
			
			>>> revolution(
			... 	ArcThrough(4*Z+Y, 6*Z, 4*Z-Y), 
			... 	Axis(O,Y), 
			... 	1.5*pi,
			... 	)
	'''
	if not isinstance(shape, (Mesh,Web,Wire)):	
		shape = web(shape)
	# get the maximum radius, to compute the curve resolution
	radius = 0
	for pt in shape.points:
		radius = max(radius, length(noproject(pt-axis[0], axis[1])))
	div = settings.curve_resolution(abs(angle*radius), abs(angle), resolution)
	def links():
		for i in range(div):          yield (i,i+1, 0)
		if abs(angle-2*pi) <= NUMPREC:    yield (div, 0, 0)
		else:                             yield (div, div+1, 0)
	return extrans(shape, (
		rotatearound(t*angle, axis)
		for t in linrange(0-alignment, 1-alignment, div=div)
		), links())

def helix(shape, height:float, angle:float, radius:float=1., axis=Axis(O,Z), alignment:float=0., resolution=None) -> Mesh:
	''' Extrude the given shape by rotating and translating along an axis
		
		This variant expects the input shape to be close to orthogonal to the axis direction and is used to produce an screw/helix from its section
		
		Parameters:
			shape:    the shape to extrude (Web, Wire, or Mesh), ideally a section
			height:   the maximum translation in the `axis` direction
			radius:   the radius at which the hexlix `angle` is computed
			angle:    the helix angle at `radius`
			axis:     the axis to rotate around and translate along
			alignment:  the relative position of the input shape in the resulting mesh
				- `0` means at the beginning
				- `1` means at the end
			resolution:   resolution setting for subdivisions
			
		Example:
			
			>>> helix(
			... 	regon(Axis(O,Z), 1, 4).subdivide(4), 
			... 	height=2, 
			... 	angle=radians(45),
			... 	)
	'''
	helix = tan(angle)
	div = settings.curve_resolution(height*helix, height/radius*helix, resolution)
	return extrans(shape, (
		translate(t*height*axis[1]) * rotatearound(t*height*helix/radius, axis)
		for t in linrange(0-alignment, 1-alignment, div=div)
		))

def screw(shape, turns:float=1., axis=Axis(O,Z), step:float=None, alignment:float=0., resolution=None):
	''' Extrude the given shape by rotating and translating along an axis
	
		This variant expexts the input shape to be close to coplanar to the axis direction and is used to produce a screw/helix from its profile
	
		Parameters:
			shape:    the shape to extrude (Web, Wire, or Mesh), ideally a profile
			turns:    number of complete rotations of the profile
			step:     the translation after a complete rotation, if not provided it is automatically adjusted to the profile's height
			axis:     the axis to rotate around and translate along
			alignment:  the relative position of the input shape in the resulting mesh
				- `0` means at the beginning
				- `1` means at the end
			resolution:   resolution setting for subdivisions
			
		Example:
			
			>>> screw(
			... 	wire([vec3(0,1,1), vec3(0,2,1), vec3(0,1,0)]).segmented(),
			... 	turns=2,
			... 	)
	'''
	if not isinstance(shape, (Mesh,Web,Wire)):	
		shape = web(shape)
	angle = turns * 2*pi
	# get the maximum radius, to compute the curve resolution
	# and the step if not provided
	radius = 0
	zmin, zmax = inf, -inf
	for pt in shape.points:
		z = dot(pt-axis[0], axis[1])
		v = noproject(pt-axis[0], axis[1])
		zmin = min(zmin, z)
		zmax = max(zmax, z)
		radius = max(radius, length(v))
	if step is None:
		step = zmax - zmin
	# produce the mesh
	div = settings.curve_resolution(abs(angle*radius), abs(angle), resolution)
	return extrans(shape, (
		translate(t*step*turns*axis[1]) * rotatearound(t*angle, axis)
		for t in linrange(0-alignment, 1-alignment, div=div)
		))

def saddle(a, b:Web) -> Mesh:
	''' Create a surface by extruding outine1 translating each instance to the next point of outline2
		
		Example:
		
		>>> saddle(
		... 	ArcThrough(+Y,X,-Y),
		... 	Softened([0*X, 0*X-2*Z, 4*X-2*Z, 4*X]),
		... 	)
	'''
	if not isinstance(a, (Mesh,Web,Wire)):
		a = web(a)
	b = web(b)
	def trans():
		s = b.points[0]
		for p in b.points:
			yield translate(mat4(1), p-s)
	return extrans(a, trans(), ((*e,t)  for e,t in zip(b.edges, b.tracks)))

def tube(shape, path:Wire, end=True, section=True) -> Mesh:
	''' Create a tube surface by extruding the shape along the path if `section` is True, there is a correction of the segments to keep the section rigid by the curve
	
		Example:
			
		>>> tube(
		... 	ArcThrough(+Y,X,-Y),
		... 	Softened([0*X, 0*X-2*Z, 4*X-2*Z, 4*X]),
		... 	)
	'''
	path = wire(path)
	def trans():
		lastrot = quat()
		l = len(path)-1
		yield mat4(1)
		for i in range(1,l+1):
			
			if i < l:
				# calculer l'angle a ce coude
				o = path[i]
				v1 = normalize(path[i-1]-o)
				v2 = normalize(path[i+1]-o)
				c = cross(-v1,v2)
				o = dot(-v1,v2)
				cl = length(c)
				cn = c/cl
				ha = atan2(cl,o)/2
				hrot = angleAxis(ha, cn)
				# calcul de la rotation de la section
				rot = hrot * lastrot
				lastrot = hrot * rot
				m = mat3_cast(rot)
				# deformation de la section pour ne pas réduire le volume
				if section:	
					m = scaledir(normalize(v1+v2), 1/cos(ha)) * m
			else:
				m = mat3_cast(lastrot)
			# construction de la transformation
			yield transform(path[i]) * mat4(m) * transform(-path[0])
	
	trans = trans()
	if not end:		next(trans)
	# handle wires with no tracks
	if path.tracks:		links = ((i, i+1, path.tracks[i]) for i in range(len(path)-1))
	else:				links = ((i, i+1, 0) for i in range(len(path)-1))
	# generate
	return extrans(shape, trans, links)

def extrans(section, transformations:iter, links=None) -> Mesh:
	''' Create a surface by extruding and transforming the given outline.
		
		Parameters:
			section:           a `Web` or a `Mesh`
			transformations:   iterable of mat4, one each section
			link:              iterable of tuples (a,b,t)  with:
									`(a,b)` the sections to link (indices of values returned by `transformation`).
									`t` the group number of that link, to combine with the section groups		
									
									if `links` is not specified, it will link each transformed section to the previous one.
									This is equivalent to giving links `(i, i+1, 0)`
	
		Example:
			
			>>> extrans(
			... 	regon(Axis(O,Z), 1, 4), 
			... 	[translate(t*Z) * scale(vec3(0.5+t**2))  for t in linrange(-1, 1, div=10)],
			... 	)
	'''
	# prepare
	if isinstance(section, Mesh):
		face = copy(section)
		face.strippoints()
		section = face.outlines()
	else:
		face = None
		section = copy(web(section))
	reindex = section.strippoints()
	
	mesh = Mesh()	  # result mesh
	extremities = {}  # index of extremities in links graph (similar to Web extremities), associated to a boolean telling the direction of that extremity
	kept = {}         # transformations kept in memory, indexed by their sections
	groups = {}       # groups created by the extransion
	
	l = len(section.points)
		
	# generate all sections points using transformations
	k = 0
	for k,trans in enumerate(transformations):
		for p in section.points:
			mesh.points.append(vec3(trans*vec4(p,1)))
		if k and not links:
			al, bl, u = (k-1)*l, k*l, 0
			for (c,d),v in zip(section.edges, section.tracks):
				t = groups.setdefault((u,v), len(groups))
				mkquad(mesh, (al+c, al+d, bl+d, bl+c), t)
		# keep extremities transformations
		elif face:
			kept[k] = trans
	
	# generate all sections faces using links
	if links:
		for a,b,u in links:
			al = a*l
			bl = b*l
			for (c,d),v in zip(section.edges, section.tracks):
				t = groups.setdefault((u,v), len(groups))
				mkquad(mesh, (al+c, al+d, bl+d, bl+c), t)
			# find extremities
			if face:
				if a in extremities:   del extremities[a]
				else:	               extremities[a] = True
				if b in extremities:   del extremities[b]
				else:                  extremities[b] = False
			
	if not links:
		extremities[0] = True
		extremities[k] = False
	
	# generate all combined groups
	mesh.groups = [None] * len(groups)
	for (u,v), t in groups.items():
		mesh.groups[t] = section.groups[v]	# NOTE will change in the future to mention both u and v
	
	# append faces at extremities
	if face:
		merges = {}  # point merge dictionnary at faces insertion
		
		for k, orient in extremities.items():
			for src,dst in enumerate(reindex):
				if dst >= 0:
					merges[src + len(mesh.points)] = dst + l*k
			end = face .transform(kept[k])
			mesh += end if extremities[k] else end.flip()
		mesh.mergepoints(merges)
	
	return mesh
	


# --- filling things ---

def flatsurface(outline, normal=None) -> 'Mesh':
	''' Generates a surface for a flat outline using the prefered triangulation method .
	
		if `normal` is specified, it must be the normal vector to the plane, and will be used to orient the face.
	'''
	if isinstance(outline, Wire):
		m = triangulation.triangulation_outline(outline, normal)
	else:
		m = triangulation.triangulation(web(outline), normal)
	if normal and dot(m.facenormal(0), normal) < 0:
		m = m.flip()
	return m


def icosurface(pts, ptangents, resolution=None) -> 'Mesh':
	''' Generate a surface ICO (a subdivided triangle) with its points interpolated using interpol2tri.
	
		- If normals are given instead of point tangents (for ptangents), the surface will fit a sphere.
		- Else `ptangents` must be a list of couples (2 edge tangents each point).
	'''
	# compute normals to points
	if isinstance(ptangents[0], tuple):
		normals = [None]*3
		for i in range(3):
			normals[i] = normalize(cross(ptangents[i][0], ptangents[i][1]))
	else:
		# if normals are given instead of tangents, compute tangents to fit a sphere surface
		normals = ptangents
		ptangents = [None]*3
		for i in range(3):
			ptangents[i] = (
				normalize(noproject(pts[i-2]-pts[i], normals[i])) * arclength(pts[i], pts[i-2], normals[i], normals[i-2]),
				normalize(noproject(pts[i-1]-pts[i], normals[i])) * arclength(pts[i], pts[i-1], normals[i], normals[i-1]),
				)
	
	# evaluate resolution (using a bad approximation of the length for now)
	div = max(( settings.curve_resolution(
					distance(pts[i-1], pts[i-2]), 
					anglebt(normals[i-1], normals[i-2]), 
					resolution)
				for i in range(3) ))
	
	return dividedtriangle(lambda u,v: intri_sphere(pts, ptangents, u,v), div)
	
def dividedtriangle(placement, div=1) -> 'Mesh':
	''' Generate a subdivided triangle with points placed according to the placement closure
		`placement(a,b) -> vec3`
		with a,b such as a+b+c = 1 with a,b,c within [0;1]
	'''
	segts = div+2
	# place points
	mesh = Mesh(groups=[None])
	for i in range(segts):
		u = i/(segts-1)				
		for j in range(segts-i):
			v = j/(segts-1)
			p = placement(u,v)
			mesh.points.append(p)
	# create faces
	c = 0
	for i in reversed(range(1,segts+1)):
		for j in range(i-1):
			s = c+j
			mesh.faces.append((s, s+i, s+1))
		for j in range(1,i-1):
			s = c+j
			mesh.faces.append((s, s+i-1, s+i))
		c += i
	mesh.tracks = typedlist.full(0, len(mesh.faces), 'I')

	return mesh


# --- standard shapes ---
	
def brick(*args, **kwargs) -> 'Mesh':
	''' A simple brick with rectangular axis-aligned sides 
	
		It can be constructed in the following ways:
		
			- brick(Box)
			- brick(min, max)
			- brick(center=vec3(0), width=vec3(-inf))
			
		Parameters:
			
			min:	the corner with minimal coordinates
			max:	the corner with maximal coordinates
			center: the center of the box
			width:  the all positive diagonal of the box
	'''
	if len(args) == 1 and not kwargs and isinstance(args[0], Box):		
		box = args[0]
	else:							
		box = Box(*args, **kwargs)
	mesh = Mesh(
		[
			vec3(1, 0, 0),
			vec3(1, 0, 1),
			vec3(0, 0, 1),
			vec3(0, 0, 0),
			vec3(1, 1, 0),
			vec3(1, 1, 1),
			vec3(0, 1, 1),
			vec3(0, 1, 0)],
		[
			uvec3(0, 1, 2),
			uvec3(0, 2, 3),
			uvec3(4, 7, 6),
			uvec3(4, 6, 5),
			uvec3(0, 4, 5),
			uvec3(0, 5, 1),
			uvec3(1, 5, 6),
			uvec3(1, 6, 2),
			uvec3(2, 6, 7),
			uvec3(2, 7, 3),
			uvec3(4, 0, 3),
			uvec3(4, 3, 7)],
		[	0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5	],
		[None] * 6,
		)
	for i in range(len(mesh.points)):
		mesh.points[i] = mesh.points[i]*box.width + box.min
	return mesh
	
def parallelogram(*directions, origin=vec3(0), align=vec3(0), fill=True) -> 'Mesh':
	''' Create a parallelogram or parallelepiped depending on the number of directions given
	
		Parameters:
			
			directions:	list of 1-3 directions, they must for a right handed base for the face normals to be oriented outward
			origin: origin the resulting shape, the shape will placed relatively to that point
			align: relative position of the origin in the shape: 0 means at start of each direction, 1 means at the tip of each direction
			fill: 
				- if True, a mesh will be generated (forming a surface with 2 directions, or an envelope with 3 directions)
				- if False, a Web will be generated
	'''
	try:				align = iter(align)
	except TypeError:	align = itertools.repeat(align)
	
	# generate points by binary combinations
	points = []
	min = origin - sum(a*d  for a,d in zip(align, directions))
	for i in range(2**len(directions)):
		points.append(min + sum(d if i>>k & 1 else 0   for k,d in enumerate(directions)))
	
	# mesh
	if len(directions) == 1:
		if fill:
			raise ValueError('cannot fill parallelogram with one only direction')
		else:
			return Web(points, [uvec2(0,1)])
	
	if len(directions) == 2:
		if fill:
			return Mesh(points, [
					uvec3(0,1,2), uvec3(2,1,3),
					])
		else:
			return Web(points, [
					uvec2(0,1),
					uvec2(1,3),
					uvec2(3,2),
					uvec2(2,0),
					]).segmented()
					
	elif len(directions) == 3:
		if fill:
			return Mesh(points, 
					[
						uvec3(0,2,1), uvec3(1,2,3),
						uvec3(0,1,4), uvec3(1,5,4),
						uvec3(0,4,2), uvec3(2,4,6),
						uvec3(4,5,6), uvec3(5,7,6),
						uvec3(2,6,3), uvec3(3,6,7),
						uvec3(1,3,5), uvec3(3,7,5),
					],
					[
						0, 0,
						1, 1,
						2, 2,
						3, 3,
						4, 4,
						5, 5,
					])
		else:
			return Web(points, [
					uvec2(0,1), uvec2(2,3), uvec2(4,5), uvec2(6,7),
					uvec2(0,2), uvec2(1,3), uvec2(4,6), uvec2(5,7),
					uvec2(0,4), uvec2(1,5), uvec2(2,6), uvec2(3,7),
					]).segmented()
		
	else:
		raise ValueError('wrong number of directions')
	
def cylinder(bottom:vec3, top:vec3, radius:float, fill=True, resolution=None) -> 'Mesh':
	''' Create a revolution cylinder, with the given radius 
	
		Parameters:
		
			bottom, top (vec3): the cylinder extremities centers
			fill (bool):        whether to put faces at both extremities
	'''
	direction = top-bottom
	base = wire(primitives.Circle(Axis(bottom,normalize(direction)), radius), resolution=resolution)
	if fill:	base = flatsurface(base).flip()
	return extrusion(base, direction)

def cone(summit:vec3, base:vec3, radius:float, fill=True, resolution=None) -> 'Mesh':
	''' Create a revolution cone, with a base of the given radius 
	
		Parameters:
			
			summit (vec3):  The point at the top of the cone
			base (vec3):    the center point of the base
			fill (bool):    whether to put a face at the base
	'''
	base = wire(primitives.Circle(Axis(base, normalize(summit-base)), radius), resolution=resolution)
	if fill:	base = flatsurface(base)
	return pyramid(summit, base)
		
def pyramid(summit:vec3, base) -> 'Mesh':
	''' Create a pyramid with the given summit point and the given base 
	
		Parameters:
			summit (vec3):   the top (summit) of the cone, not necessarity in the center of the shape
			base: (Mesh,Web,Wire):  the base shape
	'''
	if isinstance(base, Mesh):
		outline = base.outlines().flip()
		outline.stripgroups()
		result = Mesh(base.points, groups=outline.groups)
	else:
		outline = web(base)
		result = Mesh(points=outline.points, groups=outline.groups)
	
	p = len(result.points)
	result.points.append(summit)
	for edge, track in zip(outline.edges, outline.tracks):
		result.faces.append(uvec3(edge, p))
		result.tracks.append(track)
	
	if isinstance(base, Mesh):
		result += base.flip()

	return result

def square(axis:primitives.Axis, width:float) -> 'Mesh':
	''' Return a simple square with the given normal axis and square width.
		Useful to quickly create a cutplane
	'''
	x,y,z = dirbase(axis[1])
	return Mesh(
		typedlist([axis[0]+width*p   for p in ((x+y), (y-x), (-y-x), (-y+x))]),
		typedlist([uvec3(0,1,2), uvec3(2,3,0)]),
		groups=[None],
		)

def icosahedron(center:vec3, radius:float) -> 'Mesh':
	''' A simple icosahedron (see https://en.wikipedia.org/wiki/Icosahedron) '''
	phi = (1+ sqrt(5)) /2	# golden ratio
	m = Mesh(
		typedlist([
			vec3(0, 1, phi),
			vec3(1, phi, 0),
			vec3(phi, 0, 1),
			vec3(0, -1, phi),
			vec3(-1, phi, 0),
			vec3(phi, 0, -1),
			vec3(0, 1, -phi),
			vec3(1, -phi, 0),
			vec3(-phi, 0, 1),
			vec3(0, -1, -phi),
			vec3(-1, -phi, 0),
			vec3(-phi, 0, -1),
		]),
		typedlist([
			uvec3(0,1,4), uvec3(0,2,1), uvec3(0,3,2), uvec3(0,8,3), uvec3(0,4,8),
			uvec3(2,5,1), uvec3(1,6,4), uvec3(4,11,8), uvec3(8,10,3), uvec3(3,7,2),
			uvec3(3,10,7), uvec3(2,7,5), uvec3(1,5,6), uvec3(4,6,11), uvec3(8,11,10),
			uvec3(7,9,5), uvec3(5,9,6), uvec3(6,9,11), uvec3(11,9,10), uvec3(10,9,7),
		]),
		)
	f = radius/length(m.points[0])
	for i,p in enumerate(m.points):
		m.points[i] = f*p + center
	return m

def icosphere(center:vec3, radius:float, resolution=None) -> 'Mesh':
	''' A simple icosphere with an arbitrary resolution (see https://en.wikipedia.org/wiki/Geodesic_polyhedron).
	
		Points are obtained from a subdivided icosahedron and reprojected on the desired radius.
	'''
	div = settings.curve_resolution(2/6*pi*radius, 2/6*pi, resolution)
	ico = icosahedron(center, radius).subdivide(div-1)
	for i,p in enumerate(ico.points):
		ico.points[i] = center + radius * normalize(p-center)
	return ico

def uvsphere(center:vec3, radius:float, alignment=vec3(0,0,1), resolution=None) -> 'Mesh':
	''' A simple uvsphere (simple sphere obtained with a revolution of an arc) '''
	x,y,z = dirbase(alignment)
	mesh = revolution(
			web(primitives.ArcCentered(
				(center,x), 
				center+radius*z, 
				center-radius*z, 
				resolution=resolution)),
			Axis(center, z),
			resolution=resolution)
	mesh.mergeclose()
	return mesh

def regon(axis:primitives.Axis, radius, n, alignment=vec3(1,0,0)) -> 'Wire':
	''' Create a regular n-gon `Wire`, the same way we create a `Circle` '''
	x,y,z = dirbase(axis[1], alignment)
	return wire(typedlist(
		axis[0] + radius*(cos(2*pi*i/n)*x + sin(2*pi*i/n)*y)  
		for i in range(n)
		)).close().segmented()

def repeat(pattern, repetitions:int, transform):
	''' Create a mesh duplicating n times the given pattern, each time applying the given transform.
		
		Parameters:
		
			pattern:   can either be a `Mesh`, `Web` or `Wire`   
						the return type will depend on the input type
			repetitions:   the number of repetitions
			transform:     is the transformation between each duplicate
	'''
	current = pattern
	pool = type(pattern)(groups=pattern.groups)
	if repetitions:	pool += current
	for i in range(1, repetitions):
		current = current.transform(transform)
		pool += current
	return pool

def repeataround(pattern, repetitions:int=None, axis=Axis(O,Z), angle=2*pi):
	''' same as [repeat] using [rotatearound] '''
	if repetitions is None:
		if isinstance(pattern, Mesh):	indices = (i for f in pattern.faces for i in f )
		elif isinstance(pattern, Web):	indices = (i for e in pattern.edges for i in e)
		elif isinstance(pattern, Wire): indices = iter(pattern.indices)

		x,y,z = dirbase(axis[1], align=pattern.points[next(indices)] - axis[0])
		lower, upper = 0, 0
		for p in indices:
			t = atan2(dot(pattern.points[p]-axis[0], y), dot(pattern.points[p]-axis[0], x))
			lower = min(lower, t)
			upper = max(upper, t)
		repetitions = round(angle / (upper - lower))
	return repeat(pattern, repetitions, rotatearound(angle/repetitions, axis))




# This file is part of pymadcad,  distributed under license LGPL v3

'''	
	This modules provide boolean operations for all kind of meshes.
	
	Strictly speaking, boolean operations are operations on mathematically defined sets. In a n-dimensinal space it applies to subsets of point of the same dimension (volumes in 3D, surfaces in 2D).
	Here, volumes are defined by their exterior envelope (a Mesh) and surfaces by their contour (Web). So the boolean operation consist of intersecting the envelopes and contours and decide which cutted parts to keep.
	
	.. image:: /schemes/boolean-principle.svg

	This relies on a new intersection algorithm named syandana. It finds candidates for intersections using a spacial hashing of triangles over a voxel (see madcad.hashing). This is solving the problem of putting triangles in an octree.
	Also to avoid the increasing complexity of the operation with flat planes divided in multiple parallel triangles, the algorithm is implemented with a detection of ngons.
	
	The syandana algorithm achieves intersections of meshes in nearly `O(n)` where usual methods are `O(n**2)`
	
	After intersection, the selection of surface sides to keep or not is done through a propagation.

	
	Note:
	
		Web boolean operations theoretically means something only in a 2d space, and it takes place in a 3d space here, causing many situations where a web portion can be both inside and outside the contour. The algorithm here works even with meshes that are not planar at all. but it always assumes that the web you give it can be processed consistently as if it was.
	
	Note:
		
		Mesh boolean operation is relying on the surface outer normal to decide what is exterior and interior. It's always a local decision (ie. a decision made at the intersection), So the meshes must be well oriented.
		
		Web boolean operations on the other hand CANNOT be local due to the 3d space it's laying in. (a higher dimension than the surface it's theoreticall meant to outline). So it may sometimes not behave the way you expect it.
		To solve ambiguities of interior and exterior, the most outer part of first mesh argument is always considered to belong to the exterior. And the information is propagated through the web.
'''

from copy import copy, deepcopy
from operator import itemgetter
from functools import singledispatch
from time import time
from math import inf
from .mathutils import *
from .hashing import *
from . import core
from .mesh import Mesh, Web, Wire, web, line_simplification
from . import triangulation

__all__ = ['pierce', 'boolean', 'intersection', 'union', 'difference']



	
# ------- implementated operators -------



def cut_mesh(m1, m2, prec=None) -> '(Mesh, Web)':
	''' Cut m1 faces at their intersections with m2. 
		
		Return:	
			
			`(cutted, frontier)`
			
				:cutted(Mesh):	is `m1` with the faces cutted, but representing the same surface as before
				:frontier(Web): is a Web where edges are the intersection edges, and whose groups are the causing faces in `m2`
		
		Returning the intersection edges in m1 and associated m2 faces.
		The algorithm is using ngon intersections and retriangulation, in order to avoid infinite loops and intermediate triangles.
	'''
	if not prec:	prec = m1.precision()
	frontier = Web(m1.points, groups=m2.faces)	# cut points for each face from m2
	
	# topology informations for optimization
	points = PointSet(prec, manage=m1.points)
	prox2 = PositionMap(max(meshcellsize(m2), meshcellsize(m1)))
	for f2 in range(len(m2.faces)):
		prox2.add(m2.facepoints(f2), f2)
	conn = connef(m1.faces)
	
	mn = Mesh(m1.points, groups=m1.groups)	# resulting mesh
	grp = [-1]*len(m1.faces)	# flat region id
	currentgrp = -1
	for i in range(len(m1.faces)):
		# process the flat surface starting here, if the m1's triangle hits m2
		if grp[i] == -1 and m1.facepoints(i) in prox2:
			
			# a triangle cutted by an other will split it into 5 triangles, and each will be divided in 5 more if the cutting is stupidly incrmental
			# here we collect all the planar triangles and cut it as one n-gon to reduce the complexity
			# get the flat region - aka the n-gon to be processed
			currentgrp += 1
			surf = []
			front = [i]
			normal = m1.facenormal(i)
			if not isfinite(normal):	continue
			track = m1.tracks[i]
			while front:
				fi = front.pop()
				if grp[fi] == -1 and m1.tracks[fi] == track and distance2(m1.facenormal(fi), normal) <= NUMPREC:
					surf.append(fi)
					grp[fi] = currentgrp
					f = m1.faces[fi]
					for edge in ((f[1],f[0]), (f[2],f[1]), (f[0],f[2])):
						if edge in conn:	front.append(conn[edge])
			
			# get region ORIENTED outlines - aka the outline of the n-gon
			outline = set()
			for f1 in surf:
				f = m1.faces[f1]
				for edge in ((f[1],f[0]), (f[2],f[1]), (f[0],f[2])):
					if edge in outline:	outline.remove(edge)
					else:				outline.add((edge[1], edge[0]))
			original = set(outline)
			
			# process all ngon triangles
			# enrich outlines with intersections
			segts = {}
			for f1 in surf:
				f = m1.faces[f1]
				for f2 in set( prox2.get(m1.facepoints(f1)) ):
					intersect = core.intersect_triangles(m1.facepoints(f1), m2.facepoints(f2), 8*prec)
					if intersect:
						ia, ib = intersect[0][2], intersect[1][2]
						if distance2(ia, ib) <= prec**2:	continue
						# insert intersection points
						seg = (points.add(ia), points.add(ib))
						# associate the intersection edge with the m2's face
						if seg in segts:	continue
						segts[seg] = f2
						
						# cut the outline if needed
						for i in range(2):
							ii = intersect[i]
							if ii[0] != 0:	continue
							o = f[ii[1]], f[ii[1]-2]
							if o not in original:	continue
							# find the place where the outline is cutted
							p = m1.points[seg[i]]
							e = min(outline, key=lambda e: distance_pe(p, (m1.points[e[0]], m1.points[e[1]])))  # this perfect minimul should not be needed as the first below the precision should be unique, but for precision robustness ...
							if seg[i] not in e:
								#print('  cross', i, e)
								outline.remove(e)
								outline.add((e[0],seg[i]))
								outline.add((seg[i],e[1]))
										
			# simplify the intersection lines
			segts = Web(m1.points, segts.keys(), segts.values(), frontier.groups)
			segts.mergepoints(line_simplification(segts, prec))
			frontier += segts
			
			# retriangulate the cutted surface
			segts.edges.extend(uvec2(b,a) for a,b in segts.edges[:])
			segts.edges.extend(outline)
			flat = triangulation.triangulation_closest(segts, normal, prec)
			# append the triangulated face, in association with the original track
			flat.tracks = typedlist.full(track, len(flat.faces), 'I')
			flat.groups = m1.groups
			
			mn += flat
	
	# append non-intersected faces
	for f,t,grp in zip(m1.faces, m1.tracks, grp):
		if grp == -1:
			mn.faces.append(f)
			mn.tracks.append(t)
	
	return mn, frontier


def pierce_mesh(m1, m2, side=False, prec=None, strict=False) -> Mesh:

	if not prec:	prec = m1.precision()
	m1, frontier = cut_mesh(m1, m2, prec)
	
	conn1 = connef(m1.faces)		# connectivity for propagation
	stops = set(edgekey(*e) for e in frontier.edges)  # propagation stop points
	
	used = [0] * len(m1.faces)	# whether to keep the matching faces
	front = []
	
	# get front and mark frontier faces as used
	for e,f2 in zip(frontier.edges, frontier.tracks):
		for edge in ((e[0],e[1]), (e[1],e[0])):
			if edge in conn1:
				fi = conn1[edge]
				f = m1.faces[fi]
				# find from which side to propagate
				for i in range(3):
					if f[i] not in edge:
						proj = dot(m1.points[f[i]] - m1.points[f[i-1]], m2.facenormal(f2))  * (-1 if side else 1)
						if proj > prec:
							used[fi] = 1
							conn1[(f[i-1], f[i])] = fi
							conn1[(f[i], f[i-2])] = fi
							front.append((f[i], f[i-1]))
							front.append((f[i-2], f[i]))
						elif proj < -prec:
							used[fi] = -1
							conn1[(f[i-1], f[i])] = fi
							conn1[(f[i], f[i-2])] = fi
							front.append((f[i], f[i-1]))
							front.append((f[i-2], f[i]))
						break
	
	if not front:
		if side:
			m1.faces = typedlist(dtype=uvec3)
			m1.tracks = typedlist(dtype='I')
		return m1
	
	## display frontier
	#if debug_propagation:
		#from . import text
		#for edge in stops:
			#p = (m1.points[edge[0]] + m1.points[edge[1]]) /2
			#scn3D.append(text.Text(p, str(edge), 9, (1, 1, 0)))
		#from .mesh import Web
		#w = Web([1.01*p for p in m1.points], frontier.edges)
		#w.options['color'] = (1,0.9,0.2)
		#scn3D.append(w)
		#scn3D.append([text.Text(p, str(i))  for i,p in enumerate(m1.points)])
	
	# propagation
	front = [e for e in front if edgekey(*e) not in stops]
	while front:
		newfront = []
		for edge in front:
			if edge in conn1:
				source = conn1[(edge[1], edge[0])]
				assert used[source]
				fi = conn1[edge]

				if not used[fi] or (used[source]*used[fi] < 0 and abs(used[fi]) > abs(used[source])):
					assert not (strict and used[fi]), "the pierced surface is both side of the piercing surface"
					used[fi] = used[source] + (1 if used[source]>0 else -1)
					f = m1.faces[fi]
					for i in range(3):
						if edgekey(f[i-1],f[i]) not in stops:
							# the connectivity might have a different face for this edge in case of empty triangles, so avoid pushing an orphan edge, lets correct the connectivity
							conn1[(f[i-1], f[i])] = fi
							assert used[fi]
							newfront.append((f[i],f[i-1]))
		front = newfront
	
	## selection of faces
	#if debug_propagation:
		#from . import text
		#for i,u in enumerate(used):
			#if u:
				#p = m1.facepoints(i)
				#scn3D.append(text.Text((p[0]+p[1]+p[2])/3, str(u), 9, (1,0,1), align=('center', 'center')))
	
	# filter mesh content to keep
	return Mesh(
			m1.points,
			(f for u,f in zip(used, m1.faces) if u > 0),
			(t for u,t in zip(used, m1.tracks) if u > 0),
			m1.groups,
			)

#debug_propagation = True
#scn3D = []
			
def boolean_mesh(m1, m2, sides=(False,True), prec=None) -> Mesh:

	if not prec:	prec = max(m1.precision(), m2.precision())
	
	mc1 = pierce_mesh(m1, m2, sides[0], prec)
	mc2 = pierce_mesh(m2, m1, sides[1], prec)
	if sides[0] and not sides[1]:		mc1 = mc1.flip()
	if not sides[0] and sides[1]:		mc2 = mc2.flip()
	res = mc1 + mc2
	res.mergeclose()
	return res



def cut_web(w1: Web, ref: Web, prec=None) -> '(Web, Wire)':
	''' Cut the web edges at their intersectsions with the `ref` web
		
		Return:
			
			`(cutted, frontier)`
			
				:cutted(Web):	is `w1` with the edges cutted, but representing the same lines as before
				:frontier(Wire): is a Web where edges are the intersection edges, and whose groups are the causing faces in `ref`
		'''
	if not prec:  prec = w1.precision()
	frontier = Wire(w1.points, [], [], groups=ref.edges)
	
	# topology informations for optimization
	points = PointSet(prec, manage=w1.points)
	prox = PositionMap(meshcellsize(ref))
	for e in range(len(ref.edges)):
		prox.add(ref.edgepoints(e), e)
	conn = connpe(w1.edges)
	
	mn = Web(w1.points, groups=w1.groups)  # resulting web
	for e1 in range(len(w1.edges)):
		processed = False  # flag enabled when the edge is reconstructed
		
		# collect edges in ref that can have intersection with e1
		close = set( prox.get(w1.edgepoints(e1)) )
		if close:
			# no need to get the straight zone around because on the case of an edge, it will not grow in complexity more than the number of cut segments
			# and an edge is easy to reweb after multiple intersections
			# compute intersections
			segts = {}
			for e2 in close:
				intersect = intersect_edges(w1.edgepoints(e1), ref.edgepoints(e2), prec)
				if intersect:
					seg = points.add(intersect)
					segts.setdefault(seg, e2)
			
			# reconstruct the geometries
			if segts:
				e = w1.edges[e1]
				# reweb the cutted edge
				direction = w1.points[e[1]] - w1.points[e[0]]
				sorted_segts = sorted(segts.items(), key=lambda s: dot(w1.points[s[0]], direction))
				# append the rewebed edge in association with the original track
				frontier += Wire(
						w1.points, 
						map(itemgetter(0), sorted_segts), 
						map(itemgetter(1), sorted_segts), 
						ref.edges,
						)
				
				suite = list(map(itemgetter(0), sorted_segts))
				if suite[0] != e[0]:	suite.insert(0, e[0])
				if suite[-1] != e[-1]:	suite.append(e[-1])
				mn += web(Wire(
						w1.points, 
						suite, 
						[w1.tracks[e1]] * len(suite), 
						w1.groups,
						))
				processed = True
		
		# keep the non intersected faces as is
		if not processed:
			mn.edges.append(w1.edges[e1])
			mn.tracks.append(w1.tracks[e1])
	
	mn.check()
	return mn, frontier

from .nprint import nprint
	
def pierce_web(web, ref, side=False, prec=None) -> Web:

	if not prec:	prec = web.precision()
	web, frontier = cut_web(web, ref, prec)
	conn = connpe(web.edges)		# connectivity
	stops = set(frontier.indices)	# propagation stop points
	
	used = [0] * len(web.edges)		# whether to keep the matching edges or not
	
	# sort points by distance to a "center"
	center = web.barycenter()
	order = sorted(range(len(web.points)), key=lambda i:  distance2(web.points[i], center))
	
	# always start an island from the most exterior
	for start in order:
		if start in stops:	continue
		ei = next(conn[start], None)
		if not ei or used[ei]:	continue
		
		# propagate
		front = [(start, not side)]		# points on the propagation front
		while front:
			last, keep = front.pop()
			for ei in conn[last]:
				if used[ei]:	continue
				used[ei] = 1 if keep else 2
				direction = int(web.edges[ei][0] == last)
				current = web.edges[ei][direction]
				front.append((current, keep ^ (current in stops)))
		
	# filter web content to keep
	return Web(
				web.points,
				[e  for u,e in zip(used, web.edges) if u == 1],
				[t  for u,t in zip(used, web.tracks) if u == 1],
				web.groups,
				)
				
def boolean_web(w1, w2, sides, prec=None) -> Web:

	if not prec:	prec = max(w1.precision(), w2.precision())
	
	mc1 = pierce_web(w1, w2, sides[0], prec)
	w2, frontier = cut_web(w2, mc1, prec)
	conn2 = connpe(w2.edges)		# connectivity
	stops = set(frontier.indices)	# propagation stop points
	
	used = [False] * len(w2.edges)  # whether to keep the matching edges or not
	
	for p2, ei in zip(frontier.indices, frontier.tracks):
		front = []	# points on the propagation front
		
		# check which side to keep to respect the choices made in the first web
		e1 = mc1.edges[ei]
		direction = sides[0] ^ sides[1] ^ (distance2(w2.points[p2], mc1.points[e1[0]]) < distance2(w2.points[p2], mc1.points[e1[1]]))
		for ei in conn2[p2]:
			if w2.edges[ei][direction] == p2:
				front.append(w2.edges[ei][direction-1])
				used[ei] = True
		
		# propagate
		while front:
			last = front.pop()
			if last in stops:	continue
			
			for ei in conn2[last]:
				if used[ei]:	continue
				used[ei] = True
				direction = int(w2.edges[ei][0] == last)
				next = w2.edges[ei][direction]
				front.append(next)
					
	# filter web content to keep
	mc2 = Web(
			w2.points,
			[e  for u,e in zip(used, w2.edges) if u],
			[t  for u,t in zip(used, w2.tracks) if u],
			w2.groups,
			)
	
	if sides[0] and not sides[1]:		mc1 = mc1.flip()
	if not sides[0] and sides[1]:		mc2 = mc2.flip()
	res = mc1 + mc2
	res.mergeclose()
	return res
		

	
def intersect_edges(e1: '(vec3,vec3)', e2: '(vec3,vec3)', prec) -> vec3:
	''' Intersection between 2 segments '''
	d1 = e1[1]-e1[0]
	d2 = e2[1]-e2[0]
	g = e1[0]-e2[0]
	p = e2[0] + unproject(noproject(g, d1), d2)
	if not isfinite(p):		return
	#prec = NUMPREC * max(length2(d1), length2(d2))
	if dot(e1[0]-p, d1) > prec or dot(p-e1[1], d1) > prec:	return
	if dot(e2[0]-p, d2) > prec or dot(p-e2[1], d2) > prec:	return
	return p

	
	
def cut_web_mesh(w1: Web, ref: Mesh, prec=None) -> '(Web, Wire)':
	''' Cut the web inplace at its intersections with the `ref` web.
	
		Return:
		
			`(cutted, frontier)`
	'''
	if not prec:  prec = w1.precision()
	frontier = Wire(w1.points, [], [], groups=ref.faces)
	
	# topology informations for optimization
	points = PointSet(prec, manage=w1.points)
	prox = PositionMap(meshcellsize(ref))
	for f in range(len(ref.faces)):
		prox.add(ref.facepoints(f), f)
	conn = connpe(w1.edges)
	
	mn = Web(w1.points, groups=w1.groups)  # resulting web
	for e1 in range(len(w1.edges)):
		processed = False	# flag enabled when the edge is reconstructed
		
		# collect edges in ref that can have intersection with e1
		close = set(prox.get(w1.edgepoints(e1)))
		if close:
		
			# no need to get the straight zone around because on the case of an edge, it will not grow in complexity more than the number of cut segments
			# and an edge is easy to remesh after multiple intersections
			# compute intersections
			segts = {}
			for f2 in close:
				intersect = intersect_edge_face(w1.edgepoints(e1), ref.facepoints(f2), prec)
				if intersect:
					seg = points.add(intersect)
					segts.setdefault(seg, f2)
		
			# reconstruct the geometries
			if segts:
				e = w1.edges[e1]
				# reweb the cutted edge
				direction = w1.points[e[1]] - w1.points[e[0]]
				sorted_segts = sorted(segts.items(), key=lambda s: dot(w1.points[s[0]], direction))
				# append the rewebed edge in association with the original track
				frontier += Wire(
						w1.points, 
						list(map(itemgetter(0), sorted_segts)), 
						list(map(itemgetter(1), sorted_segts)), 
						ref.faces,
						)
				mn += web(Wire(
						w1.points, 
						[e[0]] + list(map(itemgetter(0), sorted_segts)) + [e[1]], 
						[w1.tracks[e1]] * (len(sorted_segts)+2), 
						w1.groups,
						))
				processed = True
		
		# keep the non intersected faces as is
		if not processed:
			mn.edges.append(w1.edges[e1])
			mn.tracks.append(w1.tracks[e1])
	
	return mn, frontier
	
					
def pierce_web_mesh(w1: Web, ref: Mesh, side=False, prec=None) -> Web:

	if not prec:	prec = w1.precision()
	
	w1, frontier = cut_web_mesh(w1, ref, prec)
	conn = connpe(w1.edges)			# connectivity
	stops = set(frontier.indices)	# propagation stop points
	
	used = [False] * len(w1.edges)	# whether to keep the matching edges
	
	for p1, fi in zip(frontier.indices, frontier.tracks):
		front = []	# points on the propagation front
		
		# check which side to keep
		normal = ref.facenormal(fi)
		for ei in conn[p1]:
			e1 = w1.edges[ei]
			direction = int(e1[0] == p1)
			if side ^ (dot(w1.points[e1[direction]] - w1.points[p1], normal) > 0):
				front.append(e1[direction])
				used[ei] = True
		
		# propagate
		while front:
			last = front.pop()
			if last in stops:	continue
			
			for ei in conn[last]:
				if used[ei]:	continue
				used[ei] = True
				direction = int(w1.edges[ei][0] == last)
				next = w1.edges[ei][direction]
				front.append(next)
	
	# filter web content to keep
	return Web(
			w1.points,
			[e  for u,e in zip(used, w1.edges) if u],
			[t  for u,t in zip(used, w1.tracks) if u],
			w1.groups,
			)

def intersect_edge_face(edge: '(vec3, vec3)', face: '(vec3, vec3, vec3)', prec) -> vec3:
	''' Intersection between a segment and a triangle
		In case of intersection with an edge of the triangle or an extremity of the edge, return the formed point.
	'''
	n = cross(face[1]-face[0], face[2]-face[0])
	a, b = edge
	da = dot(face[0]-a, n)
	db = dot(face[0]-b, n)
	if abs(da) <= prec:		p = a
	elif abs(db) <= prec:	p = b
	elif da * db > 0:		return None	# the segment doesn't reach the plane
	else:
		edgedir = b-a
		proj = dot(edgedir, n)
		if abs(proj) <= prec:	return None	# the segment is parallel to the plane
		p = a + da * edgedir / proj
	
	# check that the point is inside the triangle
	for i in range(3):
		if dot(face[i] - p, normalize(cross(n,face[i]-face[i-1]))) > prec:
			return None  # the point is outside the triangle
	return p



# --------------- stable API -------------------

pierce_ops = {
	(Mesh,Mesh):	pierce_mesh,
	(Web,Web):		pierce_web,
	(Web,Mesh):		pierce_web_mesh,
	}
def pierce(m, ref, side=False, prec=None):
	''' Cut a web/mesh and remove its parts considered inside the `ref` shape
		
		Overloads:
		
		.. code::
		
			pierce(Mesh, Mesh) -> Mesh
			pierce(Web, Web) -> Web
			pierce(Web, Mesh) -> Web
	
		`side` decides which part of each mesh to keep
	
		 - False keeps the exterior part (part exclusive to the other mesh)
		 - True keeps the interior part (the common part)
	'''
	op = pierce_ops.get((type(m), type(ref)))
	if not op:
		raise TypeError('pierce is not possible between {} and {}'.format(type(m).__name__, type(ref).__name__))
	return op(m, ref, side, prec)


boolean_ops = {
	(Mesh,Mesh):	boolean_mesh,
	(Web,Web):		boolean_web,
	}
def boolean(a, b, sides=(False,True), prec=None):
	''' Cut two web/mesh and keep its interior or exterior parts
	
		Overloads:
		
		.. code::
		
			boolean(Mesh, Mesh) -> Mesh
			boolean(Web, Web) -> Web
	
		`sides` decides which part of each mesh to keep
	
		 - False keeps the exterior part (part exclusive to the other mesh)
		 - True keeps the common part
	'''
	op = boolean_ops.get((type(a), type(b)))
	if not op:
		raise TypeError('boolean is not possible between {} and {}'.format(type(a), type(b)))
	return op(a, b, sides, prec)

def union(a, b) -> Mesh:
	''' Return a mesh for the union of the volumes. 
		It is a boolean with selector `(False,False)`
	'''
	return boolean(a,b, (False,False))

def intersection(a, b) -> Mesh:	
	''' Return a mesh for the common volume. 
		It is a boolean with selector `(True, True)`
	'''
	return boolean(a,b, (True,True))

def difference(a, b) -> Mesh:	
	''' Return a mesh for the volume of `a` less the common volume with `b`
		It is a boolean with selector `(False, True)`
	'''
	return boolean(a,b, (False,True))



from .container import *


class Wire(NMesh):
	''' This class defines a mesh of points, used for two purposes:
		
		- curves, wires, paths,  storing the ordered suite of point indices
		- cloud points, points extracted from other mesh, storing the unordered points indices 
		
		Most of the methods of `Wire` are intended for the first use case.
		
		conventions:
		
		- A curve is considered closed (or to be a loop) when its final index is the same as the first.
		- tracks are matching indices, giving a group each index. But for curves, `track[i]` gives a group for edge `(indices[i], indices[i+1])`

		Attributes:
			points:	    points buffer
			indices:	indices of the line's points in the buffer
			tracks:	    group index for each point in indices
						it can be used to associate groups to points or to edges (if to edges, then take care to still have as many track as indices)
			groups:	    data associated to each point (or edge)
			options:	custom informations for the entire wire
	'''
	__slots__ = 'points', 'indices', 'tracks', 'groups', 'options'
	
	# BEGIN ----- special methods -----
	
	def __init__(self, points=None, indices=None, tracks=None, groups=None, options=None):
		self.points = ensure_typedlist(points, vec3)
		self.indices = ensure_typedlist(indices if indices is not None else range(len(self.points)), 'I')
		self.tracks = tracks if tracks is not None else None
		self.groups = groups if groups is not None else [None]
		self.options = options or {}
	
	def __len__(self):	return len(self.indices)
	def __iter__(self):	return (self.points[i] for i in self.indices)
	def __getitem__(self, i):
		''' return the ith point of the wire, useful to use the wire in a same way as list of points
		
			equivalent to `self.points[self.indices[i]]` 
		'''
		if isinstance(i, Integral):		return self.points[self.indices[i]]
		elif isinstance(i, slice):	return typedlist((self.points[j] for j in self.indices[i]), dtype=vec3)
		else:						raise TypeError('item index must be int or slice')
	
	def __add__(self, other):
		''' append the indices and points of the other wire '''
		if isinstance(other, Wire):
			r = Wire(
				self.points if self.points is other.points else self.points[:], 
				self.indices[:], 
				self.tracks[:] if self.tracks else None, 
				self.groups if self.groups is other.groups else self.groups[:],
				)
			r.__iadd__(other)
			return r
		else:
			return NotImplemented
			
	def __iadd__(self, other):
		''' append the indices and points of the other wire '''
		if isinstance(other, Wire):		
			li = len(self.indices)
			
			if self.points is other.points:
				self.indices.extend(other.indices)
			else:
				lp = len(self.points)
				self.points.extend(other.points)
				self.indices.extend(i+lp  for i in other.indices)
			
			if self.groups is other.groups:
				if self.tracks or other.tracks:
					if not self.tracks:
						self.tracks = typedlist.full(0, li, 'I')
					self.tracks.extend(other.tracks or typedlist.full(0, len(other.indices), 'I'))
			else:
				lg = len(self.groups)
				self.groups.extend(other.groups)
				if not self.tracks:	
					self.tracks = typedlist.full(0, li, 'I')
				if other.tracks:
					self.tracks.extend(track+lg	for track in other.tracks)
				else:
					self.tracks.extend(typedlist.full(lg, len(other.indices), 'I'))
			return self
		else:
			return NotImplemented
			
	# END BEGIN ----- data management -----
	
	def strippoints(self):
		''' remove points that are used by no edge
			if used is provided, these points will be removed without usage verification
			
			no reindex table is returned as its generation costs more than the stripping operation
		'''
		self.points = typedlist((self.points[i]	for i in self.indices), dtype=vec3)
		self.indices = typedlist(range(len(self.points)), dtype='I')
		if self.points[-1] == self.points[0]:	
			self.points.pop()
			self.indices[-1] = 0
	
	def mergepoints(self, merges) -> 'self':
		''' merge points with the merge dictionnary {src index: dst index}
			merged points are not removed from the buffer.
		'''
		j = 0
		for i,f in enumerate(self.edges):
			f = merges.get(f, f)
			if not f == self.edges[i-1]:
				self.indices[i] = f
				if self.tracks:
					self.tracks[j] = self.tracks[i]
				j += 1
		del self.indices[j:]
		return self
	
	def mergeclose(self, limit=None):
		''' merge close points ONLY WHEN they are already linked by an edge.
			the meaning of this method is different than `Web.mergeclose()`
		'''
		if limit is None:	limit = self.precision()
		limit *= limit
		merges = {}
		for i in reversed(range(1, len(self.indices))):
			if distance2(self[i-1], self[i]) <= limit:
				merges[self.indices[i]] = self.indices[i-1]
				self.indices.pop(i)
				if self.tracks:
					self.tracks.pop(i-1)
		if distance2(self[0], self[-1]) < limit:
			merges[self.indices[-1]] = self.indices[0]
			self.indices[-1] = self.indices[0]
			if self.tracks:
				self.tracks[-1] = self.tracks[0]
		return merges
		
	# END BEGIN ----- mesh checks -----
	
	def isclosed(self):
		''' return True if the wire is closed, meaning the first and final indices are the same '''
		return self.indices[0] == self.indices[-1]
	
	def check(self):
		''' raise if the internal data are not consistent '''
		if not (isinstance(self.points, typedlist) and self.points.dtype == vec3):	raise MeshError("points must be a typedlist(dtype=vec3)")
		if not (isinstance(self.indices, typedlist) and self.indices.dtype == 'I'): 	raise MeshError("indices must be a typedlist(dtype='I')")
		if self.tracks and not (isinstance(self.tracks, typedlist) and self.tracks.dtype == 'I'): 	raise MeshError("tracks must be a typedlist(dtype='I')")
		l = len(self.points)
		for i in self.indices:
			if i >= l:	raise MeshError("some indices are greater than the number of points", i, l)
		if self.tracks:
			if len(self.indices) != len(self.tracks):	raise MeshError("tracks list doesn't match indices list length")
			if max(self.tracks) >= len(self.groups):	raise MeshError("some tracks are greater than the number of groups", max(self.tracks), len(self.groups))

	
	# END BEGIN ----- selection methods -----
	
	def groupnear(self, point: vec3) -> int:
		''' return group id if the edge the closest to the given point '''
		return self.tracks[self.edgenear(point)]
	
	def edgenear(self, point: vec3) -> int:
		''' return the index of the closest edge to the given point '''
		return min( range(len(self.indices)-1),
					key=lambda i: distance_pe(point, self.edgepoints(i)) )
	
	def group(self, groups):
		''' extract a part of the mesh corresponding to the designated groups.
			
			Groups can be be given in either the following ways:
				- a set of group indices
					
					This can be useful to combine with other functions. However it can be difficult for a user script to keep track of which index correspond to which created group
					
				- an iterable of group qualifiers
				
					This is the best way to designate groups, and is meant to be used in combination with `self.qual()`.
					This mode selects every group having all the input qualifiers
					
			Example:
			
				>>> # create a mesh with only the given groups
				>>> mesh.group({1, 3, 8, 9})   
				<Mesh ...>
				
				>>> # create a mesh with all the groups having the following qualifiers
				>>> mesh.group(['extrusion', 'arc'])   
				<Mesh ...>
		'''
		if not self.tracks:
			return self
		if isinstance(groups, set):			pass
		elif hasattr(groups, '__iter__'):	groups = set(groups)
		else:								groups = (groups,)
		indices = typedlist(dtype=self.indices.dtype)
		tracks = typedlist(dtype=self.tracks.dtype)
		for i,t in zip(self.indices, self.tracks):
			if t in groups:
				indices.append(i)
				tracks.append(t)
		return Wire(self.points, indices, tracks, self.groups, self.options)

	# END BEGIN ----- extraction methods -----
		
	def edgepoints(self, e) -> tuple:
		''' shorthand to the tuple of points forming edge e '''
		if isinstance(e, Integral):
			e = self.edge(e)
		return self.points[e[0]], self.points[e[1]]
		
	def edgedirection(self, e) -> vec3:
		''' direction of edge e '''
		if isinstance(e, Integral):	
			e = self.edge(e)
		return normalize(self.points[e[1]] - self.points[e[0]])
	
		
	def edge(self, i) -> uvec2:
		''' ith edge of the wire '''
		return uvec2(self.indices[i], self.indices[i+1])
	
	def edges(self) -> typedlist:
		''' list of successive edges of the wire '''
		return typedlist((self.edge(i)  for i in range(len(self.indices)-1)), dtype=uvec2)
	
	def length(self) -> float:
		''' curviform length of the wire (sum of all edges length) '''
		s = 0
		for i in range(1,len(self.indices)):
			s += distance(self[i-1], self[i])
		return s
		
	def surface(self) -> float:
		''' return the surface enclosed by the web if planar and is a loop (else it has no meaning) '''
		o = self.barycenter()
		s = vec3(0)
		for i in range(1, len(self)):
			s += cross(self[i-1]-o, self[i]-o)
		return length(s)
		
	def barycenter(self) -> vec3:
		''' curve barycenter '''
		if not self.indices:	return vec3(0)
		if len(self.indices) == 1:	return self.points[self.indices[0]]
		acc = vec3(0)
		tot = 0
		for i in range(1,len(self)):
			a,b = self[i-1], self[i]
			weight = distance(a,b)
			tot += weight
			acc += weight*(a+b)
		return acc / (2*tot)
			
	def normal(self) -> vec3:
		''' return an approximated normal to the curve as if it was the outline of a flat surface.
			if this is not a loop the result is undefined.
		'''
		area = vec3(0)
		c = self.barycenter()
		for i in range(1, len(self)):
			area += cross(self[i-1]-c, self[i]-c)
		return normalize(area)
	
	def vertexnormals(self, loop=False):
		''' return the opposed direction to the curvature in each point 
			this is called normal because it would be the normal to a surface whose section would be that wire
		'''
		normals = typedlist.full(vec3(0), len(self.indices))
		for i in range(len(self.indices)):
			a,b,c = self.indices[i-2], self.indices[i-1], self.indices[i]
			normals[i-1] = normalize(normalize(self.points[b]-self.points[a]) + normalize(self.points[b]-self.points[c]))
		self._make_loop_consistency(normals, loop)
		return normals
		
	def tangents(self, loop=False):
		''' return approximated tangents to the curve as if it was a surface section.
			if this is not a loop the result is undefined.
		'''
		tangents = typedlist.full(vec3(0), len(self.indices))
		for i in range(len(self.indices)):
			a,b,c = self.indices[i-2], self.indices[i-1], self.indices[i]
			tangents[i-1] = normalize(cross(self.points[b]-self.points[a], self.points[b]-self.points[c]))
		self._make_loop_consistency(tangents, loop)
		return tangents
	
	def _make_loop_consistency(self, normals, loop):
		l = len(self.indices)
		# make normals consistent if asked
		if loop:
			# get an outermost point as it is always well oriented
			dir = self[l//2] - self[0]	# WARNING: if those two points are too close the computation is misleaded
			i = max(self.indices, key=lambda i: dot(self.points[i], dir))
			# propagation reorient
			# WARNING: if there is a cusp in the curve (2 consecutive segments with opposite directions) the final result can be wrong
			for i in range(i+1, i+l):
				j = i%l
				e = normalize(self[j]-self[j-1])
				if dot(noproject(normals[j],e), noproject(normals[j-1],e)) < 0:
					normals[j] = -normals[j]
		# propagate to borders if not loop
		else:
			normals[0] = normals[1]
			normals[-1] = normals[-2]
		# propagate to erase undefined normals
		for _ in range(2):
			for i in range(l):
				if glm.any(isnan(normals[i])):	
					normals[i] = normals[i-1]
					
	
	def flip(self) -> 'Wire':
		''' reverse direction of all edges '''
		indices = deepcopy(self.indices)
		indices.reverse()
		if self.tracks:
			tracks = deepcopy(self.tracks[:-1])
			tracks.reverse()
			tracks.append(self.tracks[-1])
		else:
			tracks = None
		return Wire(self.points, indices, tracks, self.groups, self.options)
		
	def close(self) -> 'self':
		''' make a loop of the wire by appending its first point to its end '''
		if not self.isclosed():
			self.indices.append(self.indices[0])
			if self.tracks:
				self.tracks.append(self.tracks[0])
		return self
		
	def unclose(self) -> 'Self':
		''' if this wire is a loop, create a new Wire slicing this one's indices except the last index so the loop is opened
			if this wire is not a loop, return it immediately
		'''
		if self.isclosed():
			return Wire(self.points, self.indices[:-1], self.tracks, self.groups, self.options)
		return self
		
	def segmented(self, group=None) -> 'Wire':
		''' return a copy of the mesh with a group each edge 
		
			if group is specified, it will be the new definition put in each groups
		'''
		return Wire(self.points, 
					self.indices,
					typedlist(range(len(self.indices)), dtype='I'),
					[group]*len(self.indices),
					self.options,
					)
	
	def subdivide(self, div=1) -> 'Self':
		''' Subdivide all faces by the number of cuts '''
		n = div+2
		pts = typedlist(dtype=vec3)
		indices = typedlist(dtype='I')
		tracks = typedlist(dtype='I') if self.tracks else None
		c = 0
		for i in range(len(self.indices)-1):
			# place the points
			e = self.indices[i:i+2]
			o,p0 = self.edgepoints(e)
			x = p0-o
			for j in range(n):
				u = j/(n-1)	
				p = o + u*x
				pts.append(p)
			# create the faces
			indices.extend(range(c, c+n))
			c += n
			if self.tracks:
				tracks.extend([self.tracks[i]] * (len(indices)-len(tracks)))
		
		new = Wire(pts, indices, tracks, self.groups)
		new.mergeclose()
		new.check()
		return new
	
	# END BEGIN ----- ouput methods -----
	
	def display(self, scene):
		from .. import displays
		from .conversions import web
		
		w = web(self)
		w.options = self.options
		return w.display(scene)
	
	def __repr__(self):
		return '<Wire with {} points at 0x{:x}, {} indices>'.format(len(self.points), id(self.points), len(self.indices))
	
	def __str__(self):
		return 'Wire(\n  points={},\n  indices={},\n  tracks={},\n  groups={})'.format(
					reprarray(self.points, 'points'),
					reprarray(self.indices, 'indices'),
					reprarray(self.tracks, 'tracks') if self.tracks else None,
					repr(self.groups))
					
	# END




